/* The main class of the meta-model compiler.

Metropolis: Design Environment for Heterogeneus Systems.

Copyright (c) 1998-2004 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in all
copies of this software and that appropriate acknowledgments are made
to the research of the Metropolis group.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.


METROPOLIS_COPYRIGHT_VERSION_1
COPYRIGHTENDKEY


*/


package metropolis.metamodel;

import metropolis.metamodel.backends.Backend;
import metropolis.metamodel.backends.elaborator.ElaboratorBackend;
import metropolis.metamodel.backends.metamodel.MetaModelBackend;
import metropolis.metamodel.backends.promela.PromelaBackend;
import metropolis.metamodel.backends.runtimetest.RuntimeTestBackend;
import metropolis.metamodel.backends.systemc.SystemCBackend;
import metropolis.metamodel.frontend.FileLoader;
import metropolis.metamodel.frontend.MetaModelLibrary;
import metropolis.metamodel.nodetypes.CompileUnitNode;

import java.io.File;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

//////////////////////////////////////////////////////////////////////////
//// Compiler
/**
   The main class of the meta-model compiler. This class contains the
   main() method that starts the compilation of meta-model file.
   The user supplies some information about the compilation in the command
   line:
   <ul>
   <li> The class path, i.e. the list of classes where the meta-model files
   can be found.
   <li> The list of sources, i.e. the list of source classes that have to
   be parsed.
   <li> Some flags to choose the back-end to be used
   <li> Possibly additional flags to specify more details about the
   compilation
   </ul>
   The compiler loads the basic library classes, parses the sources,
   performs semantic analysis on them and finally, applies the back-end
   specified by the user.
   <p>
   Portions of this code were derived from sources developed under the
   auspices of the Ptolemy II project.

   @author   Robert Clariso
   @version  $Id: Compiler.java,v 1.50 2004/06/04 14:58:11 cxh Exp $
*/
public class Compiler implements MetaModelStaticSemanticConstants {


    ///////////////////////////////////////////////////////////////////
    ////                          public methods                   ////

    /** Main method of the syntax checker. Reads parameters from the
     *  command line and invokes the parser to check syntax
     *  @param args The arguments to use.  To see all the arguments,
     *  run:
     *  <pre>
     *  java -classpath $METRO/src metropolis.metamodel.Compiler -help 
     *  </pre>
     */
    public static void main(String[] args) {

        _processArguments(args);
        System.out.print("Loading libraries");
        _initCompiler();
        System.out.print("\nLoading sources");
        FileLoader.setMustRebuild(_mustRebuild);
        List compiledAsts = _compileSources();
        FileLoader.setMustRebuild(false);
        System.out.println("\nInvoking backend");

        /* if all ASTs of all packages are needed by backends,
           and we want them to be passed through arguments,
           then use the following segment of code.

           List rP1 = FileLoader.getCompiledSources(1);
           List rP2 = FileLoader.getCompiledSources(2);
           List rP3 = FileLoader.getCompiledSources(3);
           List rP4 = FileLoader.getCompiledSources(4);
           Hashtable pkgs = new Hashtable();
           Iterator iter = rP1.iterator();
           while (iter.hasNext()) {
           CompileUnitNode ast = (CompileUnitNode) iter.next();
           PackageDecl pdecl = (PackageDecl) ast.getProperty(PACKAGE_KEY);
           LinkedList llist;
           if (pkgs.containsKey(pdecl))
           llist = (LinkedList) pkgs.get(pdecl);
           else {
           llist = new LinkedList();
           pkgs.put(pdecl, llist);
           }
           llist.addLast(ast);
           }

           LinkedList groupedAsts = new LinkedList();

           CompileUnitNode cun = (CompileUnitNode)compiledAsts.get(0);
           PackageDecl pdecl = (PackageDecl) cun.getProperty(PACKAGE_KEY);
           groupedAsts.addLast(pkgs.get(pdecl));
           pkgs.remove(pdecl);
           groupedAsts.addLast(null);
           groupedAsts.addLast(null);

           iter = pkgs.keySet().iterator();
           while (iter.hasNext()) {
           PackageDecl pd = (PackageDecl) iter.next();
           if (pd.equals(MetaModelLibrary.LANG_PACKAGE))
           groupedAsts.set(1, pkgs.get(pd));
           else if (pd.equals(MetaModelLibrary.UTIL_PACKAGE))
           groupedAsts.set(2, pkgs.get(pd));
           else
           groupedAsts.addLast(pkgs.get(pd));
           //System.out.println(pd + " has " +
           //   ((LinkedList)pkgs.get(pd)).size()+" ASTs.");
           }

           compiledAsts = groupedAsts;
        ******************************************************/

        // FIXME: Testing template elimination visitor
        /*System.out.println("Invoking template elimination");
          CompileUnitNode ast = (CompileUnitNode) compiledAsts.get(0);
          List classes = ast.getDefTypes();
          UserTypeDeclNode aClass = (UserTypeDeclNode) classes.get(0);
          ObjectDecl decl =
          (+ObjectDecl) MetaModelDecl.getDecl(aClass.getName());
          ArrayList list = new fArrayList();

          list.add(metropolis.metamodel.nodetypes.IntTypeNode.instance);
          TreeNode node = TemplateHandler.instantiate(decl, list);
          classes.add(node);
        */

        _backend.invoke(_backendArgs, compiledAsts);
    }

    ///////////////////////////////////////////////////////////////////
    ////                        protected variables                ////

    /** List of paths where meta-model files and libraries can be
     *  found. */
    protected static final List _classPath = new LinkedList();

    /** List of source meta-model files that have to be compiled.
     *  Files that are imported or extended in these files are
     *  automatically compiled as well. */
    protected static final List _sources = new LinkedList();

    /** Back-end module that will be used after front-end. By default,
     *  use the meta-model back-end. */
    protected static Backend _backend = new MetaModelBackend();

    /** List of arguments that will be passed to the back-end. */
    protected static final List _backendArgs = new LinkedList();


    ///////////////////////////////////////////////////////////////////
    ////                         protected methods                 ////

    /** Compile the source files specified in the command-line. The
     *  options specified in the command-line will be used in the
     *  compilation.
     *  @return A list of the abstract syntax trees  produced by the
     *  front-end.
     */
    protected static List _compileSources() {
        List compiledAsts = new LinkedList();
        Iterator sourceIter = _sources.iterator();
        while (sourceIter.hasNext()) {
            String source = (String) sourceIter.next();
            CompileUnitNode ast = FileLoader.loadCompileUnit(source, 4);
            compiledAsts.add(ast);
        }
        return compiledAsts;
    }

    /** Initialize the meta-model compiler. Set the class path values
     *  used by the compiler to find meta-model files and libraries.
     */
    protected static void _initCompiler() {
        FileLoader.setClassPath(_classPath);
        MetaModelLibrary.initLibraries();
    }

    /** Print a help message. This text message is longer than the usage
     *  message, and provides a short description of all flags that can be
     *  used in the meta-model compiler.
     *  @param exit Set to true if we should exit.
     */
    protected static void _printHelp(boolean exit) {
        System.err.println("");
        for (int i = 0; i < _help.length; i++)
            System.err.println(_help[i]);
        System.err.println("");
        if (exit) {
            // We optionally exit so that we can test this method.
            System.exit(1);
        }
    }

    /** Print usage of the meta-model compiler. The text message provides
     *  brief information on how the meta-model compiler has to be used.
     *  @param exit Set to true if we should exit.
     */
    protected static void _printUsage(boolean exit) {
        System.err.println("");
        for (int i = 0; i < _usage.length; i++)
            System.err.println(_usage[i]);
        System.err.println("");
        if (exit) {
            // We optionally exit so that we can test this method.
            System.exit(1);
        }
    }

    /** Process the arguments from the command line. In case there's is
     *  something wrong with the syntax, _printUsage(true) is called, which
     *  exits the program.
     *  @param args Command-line options.
     */
    protected static void _processArguments(String[] args) {
        int numArgs    = args.length;
        int currentArg = 0;
        if (numArgs == 0) _printUsage(true);
        while (currentArg < numArgs) {
            if (args[currentArg] == null) {
                throw new RuntimeException("Compiler: Internal error: "
                        + "argument #" + currentArg + " is null.");
            }
            if (args[currentArg].charAt(0) == '-') {
                // Flag
                String current = args[currentArg];
                if (current.length() == 1) {
                    System.err.println("Error: empty flag '-' "
                            + "in command line");
                    _printUsage(true);
                }
                current = current.substring(1);
                currentArg = _processFlag(current, currentArg, numArgs, args);
            } else {
                // Source file
                do {
                    String current = args[currentArg++];
                    if (current.charAt(0) == '-') {
                        System.err.println("Error: all flags must appear"
                                + " before source files,"
                                + " arg was '" + current + "'");
                        _printUsage(true);
                    }
                    _sources.add(current);
                } while (currentArg < numArgs);
            }
        }
        if (_sources.size() == 0) {
            System.err.println("Error: source files expected");
            _printUsage(true);
        }
    }

    /** Process one flag from the command line. In case there's is
     *  something wrong with the syntax, _printUsage() is called, which
     *  exits the program.
     *  @param flag Identifier of the flag.
     *  @param pos  Position of the flag in the array of options.
     *  @param max  Total number of command-lne options.
     *  @param args Array of command-line options.
     *  @return The next command-line option to be examined.
     */
    protected static int _processFlag(String flag, int pos, int max,
            String[] args) {

        if (flag.equals("h") || flag.equals("?") || flag.equals("help")) {
            // Print a help message
            _printHelp(true);
        } else if (flag.equals("metamodel")) {
            _backendArgs.clear();
            _backend = new MetaModelBackend();
            return pos + 1;
        } else if (flag.equals("classpath")) {
            // Read all paths in the class-path list
            String file;
            do {
                // Two possible scenarios:
                // 1) path
                // 2) path1 : path2
                // If we don't find ':' after current path,
                // then it is the last one.
                if (pos + 1 >= max) {
                    System.err.println("Error: wrong classpath, path "
                            + "expected");
                    _printUsage(true);
                }
                pos++;
                file = args[pos];
                if (file.equals(":")) {
                    System.err.println("Error: ':' used where a path was"
                            + " expected");
                    _printUsage(true);
                }
                StringTokenizer tokenizer = new StringTokenizer(file,
                        File.pathSeparator);
                while (tokenizer.hasMoreTokens()) {
                    _classPath.add(tokenizer.nextToken());
                }
                if (pos + 1 >= max) {
                    System.err.println("Error: source files expected after"
                            + " class-paths");
                    _printUsage(true);
                }
                pos++;
                file = args[pos];
            } while ( file.equals(":") );
            return pos;
        } else if (flag.equals("simulator")) {
            _backendArgs.clear();
            _backend = _instantiateBackend(
                    "metropolis.metamodel.backends.simulator.SimulatorBackend");
            return pos + 1;
        } else if (flag.equals("javasim")) {
            _backendArgs.clear();
            _backend = _instantiateBackend(
                    "metropolis.metamodel.backends.javasim.JavasimBackend");
            return pos + 1;
        } else if (flag.equals("systemc")) {
            _backendArgs.clear();
            _backend = new SystemCBackend();
            return pos + 1;
        } else if (flag.equals("cfa")) {
            _backendArgs.clear();
            _backend = _instantiateBackend(
                    "metropolis.metamodel.backends.cfa.CFABackend");
            return pos + 1;
        } else if (flag.equals("cpp")) {
            _backendArgs.clear();
            _backend = _instantiateBackend(
                    "metropolis.metamodel.backends.cpp.CPPBackend");
            return pos + 1;
        } else if (flag.equals("top")) {
            if ((_backend instanceof SystemCBackend) ||
                    (_backend.getClass().getName().equals(
                            "metropolis.metamodel.backends.javasim.JavasimBackend"))) {

                if (pos == max-1) {
                    System.err.println("Error: -top must be followed by the "+
                            "fully qualified name of the net-initiator netlist");
                    _printUsage(true);
                } else {
                    _backendArgs.add(args[++pos]);
                    return pos + 1;
                }
            } else {
                return pos + 1;
            }
        } else if (flag.equals("w")) {
            _backendArgs.add(args[pos]);
            return pos + 1;
        } else if (flag.equals("java")) {
            if ((_backend instanceof SystemCBackend) ||
                    (_backend.getClass().getName().equals(
                            "metropolis.metamodel.backends.javasim.JavasimBackend"))) {
                if (pos == max-1) {
                    System.err.println("Error: -java must be followed by the "+
                            "full path of java");
                    _printUsage(true);
                } else {
                    _backendArgs.add("-java");
                    _backendArgs.add(args[++pos]);
                    return pos + 1;
                }
            } else { //For other backends which need -java, modify here.
                System.err.println("Warning: -java must appear after -systemc "
                        + "or -javasim. Ignored.");
                _printUsage(true);
            }
            return pos + 1;
        } else if (flag.equals("javac")) {
            if ((_backend instanceof SystemCBackend) ||
                    (_backend.getClass().getName().equals(
                            "metropolis.metamodel.backends.javasim.JavasimBackend"))) {
                if (pos == max-1) {
                    System.err.println("Error: -java must be followed by the "
                            + "full path of javac");
                    _printUsage(true);
                } else {
                    _backendArgs.add("-javac");
                    _backendArgs.add(args[++pos]);
                    return pos + 1;
                }
            } else { //For other backends which need -javac, modify here.
                System.err.println("Warning: -javac must appear after -systemc "
                        + " or -javasim. Ignored.");
                _printUsage(true);
            }
            return pos + 1;
        } else if (flag.equals("elaborator")) {
            _backendArgs.clear();
            _backend = new ElaboratorBackend();
            if (pos == max-1) {
                System.err.println("Error: elaborator needs the fully " +
                        "qualified name of the net-initiator netlist");
                _printUsage(true);
            } else {
                _backendArgs.add(args[++pos]);
                return pos + 1;
            }
        } else if (flag.equals("promela")) {
            _backendArgs.clear();
            _backend = new PromelaBackend();
            return pos + 1;
        } else if (flag.equals("runtimetest")) {
            _backendArgs.clear();
            _backend = new RuntimeTestBackend();
            if (pos == max-2) {
                System.err.println("Error: runtimetest needs an option and the fully "+
                        "qualified name of the net-initiator netlist");
                _printUsage(true);
            }
            else {
                _backendArgs.add(args[++pos]);
                _backendArgs.add(args[++pos]);
                return pos + 2;
            }
            //         } else if (flag.equals("prometheus")) {
            //             _backendArgs.clear();
            //             _backend = new PrometheusBackend();
            //             return pos + 1;
        } else if (flag.equalsIgnoreCase("mmdebug")) {
            if (_backend instanceof SystemCBackend) {
                _backendArgs.add(args[pos]);
                _mustRebuild = true;
                return pos + 1;
            } else {
                System.err.println("\"-mmdebug\" must appear after "
                        + "\"-systemc\", only supported for "
                        + "\"-systemc\".  Ignored.");
            }
        } else if (flag.equalsIgnoreCase("dumpast")) {
            if (_backend instanceof MetaModelBackend) {
                _backendArgs.add(args[pos]);
                return pos + 1;
            } else {
                System.err.println("\"-dumpast\" only supported for "
                        + "\"-metamodel\".  Ignored.");
            }
        } else {
            System.err.println("Error: unknown flag '-" + flag + "'");
            _printUsage(true);
        }
        return max;
    }

    ///////////////////////////////////////////////////////////////////
    ////                         private methods                   ////

    // Instantiate a backend using reflection.
    private static Backend _instantiateBackend(
            String classname) {
        Backend backend;
        try {
            Class simulatorBackendClass = Class.forName(classname);
            backend = (Backend) simulatorBackendClass.newInstance();
        } catch (Throwable throwable) {
            // FIXME: this class needs to handle exceptions better.
            throw new RuntimeException("Failed to find or instantiate '"
                    + classname + "'",
                    throwable);
        }
        return backend;
    }


    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////

    /** Usage message. This message is printed each time that this class
     *  is invoked with an unknown argument, with an end of line after
     *  each element of the array.*/
    private static final String[] _usage = {
        "Usage: java metropolis.metamodel.Compiler",
        "       [-h]  [-?]  [-help]",
        "       [-metamodel [-dumpast]]",
        "       [-simulator]",
        "       [-javasim]",
        "       [-cpp]",
        "       [-systemc [-top <netinitiator>] [-java <path>/java] [-javac <path>/javac] [-w] [-mmdebug]]",
        "       [-elaborator <netinitiator>]",
        "       [-promela]",
        // "       [-prometheus]",
        "       [-cfa]",
        "       [-classpath <classpath>]",
        "       source_files",
        "",
        "    Use 'java metropolis.metamodel.Compiler -h' for more help. "
    };

    /** Help message. This message is printed when the 'help' flag is
     *  used, with an end of line after each element of the array, */
    private static final String[] _help = {
        "class metropolis.metamodel.Compiler:",
        "       A compiler for the meta-model language, the internal representation",
        "       used in the Metropolis design framework",
        "",
        "Usage: java metropolis.metamodel.Compiler [options] source_files",
        "         ",
        "where options include:",
        "   -? -help -h                Display this help message",
        "   -classpath <classpath>     Specify where to find user class files",
        "   -metamodel [-dumpast]      Use the metamodel back-end (default)",
        "                              -dumpast prints a dump of the AST to STDOUT",
        "   -javasim [-top <netinit>] [-java <path>/java] [-javac <path>/javac]",
        "                              Generate java simulation code (semantics 3.0)",
        "                              [javasim not shipped in Metropolis 1.0]",
        "   -cpp                       Generate C++ simulation code",
        "                              [cpp not shipped in Metropolis 1.0]",
        "   -simulator                 Generate java simulation code",
        "                              [simulator not shipped in Metropolis 1.0]",
        "   -systemc [-top <netinit>] [-java <path>/java] [-javac <path>/javac] [-w] [-mmdebug]",
        "                              Generate SystemC simulation code",
        "   -elaborator <netinit>      Compute the structure of the network",
        "   -promela                   Generate Promela/Spin code",
        //"   -prometheus                Perform compositional verification",
        "   -cfa                       Generate a CFA; Visual, KISS,",
        "                              and RML representations",
        "                              [cfa not shipped in Metropolis 1.0]",
    };

    private static boolean _mustRebuild = false;


}
