/*
  Copyright (c) 2003-2004 The Regents of the University of California.
  All rights reserved.

  Permission is hereby granted, without written agreement and without
  license or royalty fees, to use, copy, modify, and distribute this
  software and its documentation for any purpose, provided that the
  above copyright notice and the following two paragraphs appear in all
  copies of this software and that appropriate acknowledgments are made
  to the research of the Metropolis group.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
  FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
  ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
  THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
  PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
  CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
  ENHANCEMENTS, OR MODIFICATIONS.


  METROPOLIS_COPYRIGHT_VERSION_1
  COPYRIGHTENDKEY

*/

package metropolis.metamodel.backends.systemc;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Vector;

//////////////////////////////////////////////////////////////////////////
//// MakefileCodeGen
/**
   Generate a makefile.

   <p>Created on May 5, 2003, 10:42 AM

   @author  Daniele Gasperini, Contributor: Christopher Brooks
   @version $Id: MakefileCodegen.java,v 1.21 2004/05/25 20:28:06 allenh Exp $
*/
public class MakefileCodegen {

    ///////////////////////////////////////////////////////////////////
    ////                         public variables                  ////

    // We use a makefile with the .mk extension so that tools can tell
    // it is a makefile
    public static String _CPP    = ".cpp";
    public static String _H      = ".h";
    public static String _O      = ".o";
    public static String _PREFIX = "MMSCS_";


    /** Create a new instance of MakefileCodegen (non-debugging version). */
    public MakefileCodegen() throws IOException {
        _file = new FileWriter(_FILENAME);
    }

    /** Create a new instance of MakefileCodegen.
     *
     * @param debuggingMetamodel <code>true</code> if generated makefile
     *	should be configured for running a debugger on the executable, with
     *	the <i>.mmm</i> files substituted for the corresponding <i>.cpp</i>
     *	files.
     */
    public MakefileCodegen(boolean debuggingMetamodel) throws IOException {
        _file = new FileWriter(_FILENAME);
	_debuggingMetamodel = debuggingMetamodel;
    }

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** Generate a header for the makefile.
     *  If a file named makefile or makefile exists, then read it
     *  and look for lines that start with "include", but do not
     *  contain $(CONFIG), metro.mk or metrocommon.mk and output them
     *  in the generated file.
     */
    public void header() throws IOException {
        // If we find a makefile or Makefile, we save the name.
        String sourceMakefile = null;

        // include lines from the makefile or Makefile
        String includes = "";
        try {
            BufferedReader bf;
            // First, we check for makefile and then for Makefile because
            // that is what GNU make does.
            try {
                sourceMakefile = "makefile";
                bf = new BufferedReader(new FileReader(sourceMakefile));

            } catch (FileNotFoundException ex) {
                sourceMakefile = "Makefile";
                bf = new BufferedReader(new FileReader(sourceMakefile));
            }

            do {
                String line = bf.readLine();
                if (line == null) {
                    break;
                }
                if (line.startsWith("include")) {
                    // Skip includes of $(CONFIG), metro.mk and metrocommon
		    if (line.indexOf("$(CONFIG)") == -1
                            && line.indexOf("metro.mk") == -1
                            && line.indexOf("metrocommon.mk") == -1 ) {
                        includes += line + "\n";
                    }
                }
            } while (true);
        } catch (Exception e) {
            // do nothing, Makefile or makefile does not exist.
        }

        _file.write("# This makefile is automatically generated by "
                + "Metropolis.\n\n");
	if (_debuggingMetamodel) {
	    _file.write("override CXX_OPTIMIZER += -ggdb\n\n");
	}
        _file.write("ROOT =      $(METRO)\n\n"
                + "CONFIG =    $(ROOT)/mk/metro.mk\n"
                + "include $(CONFIG)\n\n");
        if (includes.length() > 0) {
            _file.write("\n# The following includes appeared in the "
                    + sourceMakefile + ":\n" + includes + "\n");
        }
        _file.write("CXX_INCLUDES = " + _includes + "\n\n"
                + "LIBRARIES_PATH = " + _librariesPath + "\n\n");

        _file.write("all: bin\n\n");
    }

    /** Add a file to the list of files.
     *  @param baseFilename The file name to be added.
     */
    public void add(String baseFilename) {
        _filelist.add(baseFilename);
    }


    /** Generate the body of the makefile
     */
    public void body() throws IOException {
        for (int i = 0; i < _filelist.size(); i++) {
            String baseFilename = (String) _filelist.get(i);
            _file.write(baseFilename + _O + ": \\\n"
                    + "\t\t" + baseFilename + _CPP + " \\\n "
                    + "\t\t" + baseFilename + _H + " \n"
                    + "\t$(METRO_CXX) -c $(CXX_FLAGS) $(CXX_INCLUDES) \\\n"
                    + "\t\t" + baseFilename + _CPP + " \\\n"
                    + "\t\t -o " + baseFilename + _O + " \\\n"
                    + "\t\t" + _compFlags + "\n\n");
        }

        _file.write("bin: \\\n ");

        for (int i = 0; i < _filelist.size(); i++) {
            String baseFilename = (String) _filelist.get(i);
            _file.write("\t\t" + baseFilename + ".o" + " \\\n ");
        }

        _file.write("\t\tsc_main.o\n"
                + "\t $(METRO_CXX) $(CXX_FLAGS) $(CXX_INCLUDES) "
                + " $(LIBRARIES_PATH) -o "
                + _EXENAME + " \\\n");

        for (int i = 0; i < _filelist.size(); i++) {
            String baseFilename = (String) _filelist.get(i);
            _file.write("\t\t" + baseFilename + ".o" + " \\\n");
        }

        _file.write("\t\tsc_main.o " + _linkArchives + " \\\n "
                + "\t\t" +_linkFlags + " \\\n "
                + "\t\t" + _librariesName + "\n"
                + "\nsc_main.o: ./sc_main.cpp\n"
                + "\t$(METRO_CXX) $(CXX_FLAGS) $(CXX_INCLUDES) -c ./sc_main.cpp "
                + _compFlags + " "
                + "\n");

	if (_debuggingMetamodel) {
	    _file.write("\n"
                    + "src_substitutes: $(SRC_SUBSTITUTES)\n"
                    + "\n"
                    + "# Save .cpp files and replace them w/ .mmm files\n"
                    + "# so the debugger uses the .mmm code as the source\n#\n"
                    + "%.cpp.bak: %.cpp\n"
                    + "\tmmfile=`echo $< | sed 's/\\.cpp//'`.mmm ;\\\n"
                    + "\tif [ -f $${mmfile} ]; then \\\n"
                    // + "\t  cat $< > $@; \\\n"
                    + "\t  mv -f $< $@; \\\n"
                    + "\t  cp $${mmfile} $<; \\\n"
                    + "\tfi\n"
                    + "\n"
                    + "unsubstitute: FORCE\n"
                    + "\tfor f in $(SRC_SUBSTITUTES); do \\\n"
                    + "\t    if [ -f $$f ]; then \\\n"
                    + "\t        mv $$f `echo $$f | sed 's/\\.bak$$//'`; \\\n"
                    + "\t    fi; \\\n"
                    + "\tdone\n"
                    + "\n"
                    + "FORCE:\n\n");

	}
    }

    /** Return a string containing the include files.
     *  @return The string containing the includes.
     */
    public String createIncludes() {
        String result = "";
        for (int i = 0; i < _filelist.size(); i++) {
            result += "#include \"" + _filelist.get(i) + ".h\"\n";
        }
        return result;
    } // end createIncludes

    /** Write the last lines of the makefile and close the file descriptor */
    public void close() throws IOException {
        _file.write("\n\n# Get the rest of the rules\n"
                + "include $(ROOT)/mk/metrocommon.mk\n");
        _file.close();
    }

    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////

    private final String _EXENAME            = "run.x";
    private final String _FILENAME           = "systemc_sim.mk";

    private String       _compFlags          = "$(COMPFLAGS)";
    private boolean      _debuggingMetamodel = false;
    private FileWriter   _file               = null;
    private Vector       _filelist           = new Vector();
    private String       _includes  = "-I. $(INCDIR) "
    + "-I$(SYSTEMC_BACKEND) "
    + "-I$(SYSTEMC_BACKEND)/util "
    + "-I$(SYSTEMC_BACKEND)/ltl2ba-1.0 "
    + "-I$(SYSTEMC_BACKEND)/zchaff "
    + "-I$(SYSTEMC)/include";
    private String _librariesName = "$(LINKLIBS) -lsystemc -lm "
    + "-llibrary " + "-lutil "
    + "-lltl2ba -lsat $(LIBS)";
    private String _librariesPath = "-L. -L.. $(LIBDIR) "
    + "-L$(SYSTEMC_BACKEND)/util "
    + "-L$(SYSTEMC_BACKEND)/ltl2ba-1.0 "
    + "-L$(SYSTEMC_BACKEND)/zchaff "
    + "-L$(SYSTEMC_BACKEND) "
    + "-L$(SYSTEMC)/$(SYSTEMC_LIB)";
    private String _linkArchives  = "$(ARCHIVES)";
    private String _linkFlags     = "$(LINKFLAGS)";
}
