/* A LinkedList to hold generated SystemC code, with optional #line directives

Metropolis: Design Environment for Heterogeneus Systems.

Copyright (c) 1998-2004 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in all
copies of this software and that appropriate acknowledgments are made
to the research of the Metropolis group.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.


METROPOLIS_COPYRIGHT_VERSION_1
COPYRIGHTENDKEY
*/

package metropolis.metamodel.backends.systemc.mmdebug;

import metropolis.metamodel.MetaModelStaticSemanticConstants;
import metropolis.metamodel.NullValue;
import metropolis.metamodel.TreeNode;

import java.io.File;
import java.util.LinkedList;
import java.util.List;

//////////////////////////////////////////////////////////////////////////
//// LineList
/**
 * A <code>LinkedList</code> that knows how to add
 * "<code>#line &lt;line_number&gt; &lt;file_name&gt;</code>" preprocessor
 * directives to the generated code, taking that information from the
 * <code>TreeNode</code> that it is instantiated with.
 * <p>
 *
 * @author   Daniele Gasperini, Allen Hopkins
 * @version  $Id: LineList.java,v 1.16 2004/05/25 01:17:29 allenh Exp $
 */

public class LineList extends LinkedList
    implements MetaModelStaticSemanticConstants {

    //////////////////////////////////////////////////////////////////////////
    ///         Constructors

    /**
     * Create a non-debugging <code>LineList</code>.
     *
     * @param node The TreeNode whose generated source code this list will
     * contain.
     */
    public LineList(TreeNode node) {
	this(node, false, false);
    }

    /**
     * Create a <code>LineList</code>, with control over whether it creates
     * code for metamodel debugging.
     * <p>
     * Setting the <code>emittingLineDirectives</code> parameter to
     * <code>true</code> causes
     * "<code>#line &lt;line_number&gt; &lt;file_name&gt;</code>" preprocessor
     * directives to be added to the generated source code for use by the
     * metamodel debugger.
     *
     * @param node The TreeNode whose generated source code this list will
     * contain.
     * @param emittingLineDirectives If <code>true</code>, generated code is to
     * be used with the metamodel debugger.
     */
    public LineList(TreeNode node, boolean emittingLineDirectives) {
        this(node, emittingLineDirectives, false);
    }

    /**
     * Create a <code>LineList</code>, with control over whether it creates
     * code for metamodel debugging, and whether next-line information is
     * added automatically, or only by explicitly calling
     * <code>addBeginningLineInfo()</code>.
     * <p>
     * Setting the <code>emittingLineDirectives</code> parameter to
     * <code>true</code> causes
     * "<code>#line &lt;line_number&gt; &lt;file_name&gt;</code>" preprocessor
     * directives to be added to the generated source code for use by the
     * metamodel debugger.
     *
     * @param node The TreeNode whose generated source code this list will
     *          contain.
     * @param emittingLineDirectives If <code>true</code>, generated code is to
     *          be used with the metamodel debugger.
     * @param nextLinesAreAddedExplicitly If <code>true</code>, the user
     *          intends to call <code>addBeginningLineInfo()</code> explicitly.
     *          The default is for it to be called automatically on the
     *          first invocation of <code>addLast()</code> for this
     *          <code>node</code>.
     */
    public LineList(TreeNode node,
            boolean emittingLineDirectives,
            boolean nextLinesAreAddedExplicitly) {
        _node                 = node;
        _fileName             = (String)_node.getProperty(IDENT_KEY);
        // XXX Use short filename, until it turns out we need the full one.
        // We need the full one.
        /*
        if (_fileName != null) {
            _fileName             = DebugUtil.getNodeFileName(node);
        }
        */
        // try {
        Object lineNumberObj = _node.getProperty(LINENUMBER_KEY);
        if (lineNumberObj == NullValue.instance) {
            _lineNumber = null;
        } else {
            _lineNumber= (Integer)lineNumberObj;
        }
        // } catch (Exception ex) {
        //     throw new RuntimeException(DebugUtil.getNodeFileName(node)
        //                 + ":" + DebugUtil.getNodeKind(node), ex);
        // }
        _nextLineNumbers      = (List)_node.getProperty(NEXT_LINENUMBERS_KEY);
        _blockEndLineNumber
            = (Integer)_node.getProperty(BLOCK_END_LINENUMBER_KEY);
        _emittingLineDirectives      = emittingLineDirectives;
        _nextLinesAreAddedExplicitly = nextLinesAreAddedExplicitly;

	if (_lineNumber != null) {
	    if (_lineNumber.intValue() == -1) _lineNumber = null;
	}

        if (_fileName != null) {
            _shortFileName = _fileName.substring(
                    _fileName.lastIndexOf(File.separatorChar) + 1);
        }
    }

    //////////////////////////////////////////////////////////////////////////
    ///         public methods

    /**
     * Add some code to the list.  If we are generating code for the metamodel
     * debugger, and the string we are adding contains a newline, also add
     * debugging line-number information.
     *
     * @param o An object, which may be a <code>String</code>, or another
     * <code>List</code>.
     */
    public void addLast(Object o) {
        if (! _emittingLineDirectives
                || ! (o instanceof java.lang.String)
                || o.toString().indexOf("\n") == -1) {
	    super.addLast(o);
	} else {
	    String line = "";
            if (_nextLineNumbers != null && ! _nextLinesAreAddedExplicitly) {
                line += _getBeginningLineInfo();
            }
	    if (_fileName != null && _lineNumber != null) {
                line += "#line " + _lineNumber + " \"" + _fileName + "\"\n";
	    } else if (_fileName == null && _lineNumber != null) {
                line += "#line " + _lineNumber + "\n";
	    }
	    if (line.length() > 0) {
		super.add(_placeholder, line);
	    }
	    super.addLast(o);

            if (_methodReturnLineNumber != null &&
                    _methodReturnLineNumber != NullValue.instance) {
                super.addLast("pc->methodReturnStack.pop_back();\n");
                _methodReturnLineNumber = null;
            }

	    _placeholder = size();
        }
    }

    /**
     * Add information to the generated C++ code from this node that
     * the metamodel debugger can use to set breakpoints at the next
     * command that the current process will execute when it becomes
     * unblocked at the beginning of an await statement critical section,
     * labeled statement, or interface function call.
     */
    public void addBeginningLineInfo() {
        if (_emittingLineDirectives && _nextLineNumbers != null) {
	    String line = _getBeginningLineInfo();

            super.add(_placeholder, line);
	    _placeholder = size();
        }
    }

    /**
     * Add information to the generated C++ code from this node that
     * the metamodel debugger can use to set breakpoints at the
     * end of a labeled statement or interface function call.
     */
    public void addEndingLineInfo() {
        if (_emittingLineDirectives && _blockEndLineNumber != null) {
            StringBuffer buf = new StringBuffer();
            /***
                buf.append("pc->clearNextMMLineNumbers();");
                buf.append(_debugSayClearingLines());
                buf.append(_debugNewline());
                buf.append("pc->addNextMMLineNumber(new String(\"");
                buf.append(_fileName);
                buf.append(":");
                buf.append(_blockEndLineNumber);
                buf.append("\"));");
                buf.append(_debugSayAddingLine(_blockEndLineNumber.toString()));
            ***/

            _addCurrentLineInfo(buf);

            // Make process name, next-lines, and current-lines accessible to
            // gdb:
            /***
                buf.append("gProcessName = pc->p->name();");
                buf.append("gMmdbSB=pc->getScoreboard();");
            ***/
            /***
                buf.append("pcNextLines=pc->getNextMMLineNumbers();");
                buf.append("gAllNextLines=gMmdbSB->getAllNextLines();");
            ***/
            buf.append("gAllCurrentLines=gMmdbSB->getAllCurrentLines();\n");

            super.add(_placeholder, buf.toString());
	    _placeholder = size();
        }
    }

    /**
     * Add information to the generated C++ code from this node that
     * clears the line numbers set by the info added in addEndingLineInfo().
     */
    public void clearEndingLineInfo() {
        StringBuffer buf = new StringBuffer();
        /***
            buf.append("pc->clearNextMMLineNumbers();");
            buf.append(_debugSayClearingLines());
            buf.append(_debugSayThatsAll());
        ***/

        _addCurrentLineInfo(buf);

        buf.append("\n");

        super.add(_placeholder, buf.toString());
        _placeholder = size();
    }

    /**
     * Fix <code>pc-&gt;currentLineNumber</code> in an <code>await</code>
     * statement, after running a method in its test section may have
     * set it to something wrong.
     */
    public void setAwaitCurrentLineNumber() {
        StringBuffer buf = new StringBuffer();
        _addCurrentLineInfo(buf);
        buf.append("\n");
        super.add(_placeholder, buf.toString());
    }

    /**
     * Set the line number to be used by addEndingLineInfo().  The default
     * is to use the BLOCK_END_LINENUMBER_KEY property of the node this
     * LineList was constructed with.
     * @param blockEndLineNumber The line number to use.
     */
    public void setBlockEndLineNumber(Integer blockEndLineNumber) {
        _blockEndLineNumber = blockEndLineNumber;
    }

    /**
     * Set the line number to be used in <code>addLast()</code> to add C++
     * code to store the line number to return to after an immanent method
     * call.  This enables the debugger to set a breakpoint to stop at after
     * the method call returns.
     * @param lineNumber The line number to return to from a method call.
     */
    /***
        public void setMethodReturnLineNumber(Object lineNumber) {
        _methodReturnLineNumber = lineNumber;
        }
    ***/

    /**
     * Explictly set the line number to use in <code>#line</code> directives.
     * @param lineNumber The number to set.
     */
    public void setLineNumber(Integer lineNumber) {
        _lineNumber = lineNumber;
    }

    //////////////////////////////////////////////////////////////////////////
    ///         private methods

    private void _addCurrentLineInfo(StringBuffer buf) {
        buf.append("pc->currentLineNumber=new String(\"");
        buf.append(_shortFileName);
        buf.append(":");
        buf.append(_lineNumber);
        buf.append("\");");
        if (_DEBUGGING) {
            buf.append("gMmdbSB=pc->getScoreboard();");
            buf.append("gAllCurrentLines=gMmdbSB->getAllCurrentLines();");
            buf.append("for (int i=0;i<gAllCurrentLines.size();i++){");
            buf.append(   "cerr<<gAllCurrentLines[i]->data()<<endl;}");
            buf.append("cerr<<pc->p->name()<<endl;");
        }
    }

    private String _getBeginningLineInfo() {
        StringBuffer buf = new StringBuffer();
        /***
            int nLineNumbers = _nextLineNumbers.size();
            buf.append("pc->clearNextMMLineNumbers();");
            buf.append(_debugSayClearingLines());
            if ( nLineNumbers != 0 ) {
            buf.append(_debugNewline());
            for (int i = 0; i < nLineNumbers; i++) {
            buf.append("pc->addNextMMLineNumber(new String(\"");
            buf.append(_fileName);
            buf.append(":");
            buf.append(_nextLineNumbers.get(i));
            buf.append("\"));");
            buf.append(_debugSayAddingLine(
            _nextLineNumbers.get(i).toString()));
            }
            _addCurrentLineInfo(buf);
            } else {
            buf.append(_debugSayThatsAll());
            _addCurrentLineInfo(buf);
            }
        ***/
        _addCurrentLineInfo(buf);

        /***
            if (_methodReturnLineNumber != null &&
            _methodReturnLineNumber != NullValue.instance) {
            buf.append("pc->methodReturnStack.push_back(new String(\"");
            buf.append(_fileName);
            buf.append(":");
            buf.append(_node.getProperty(NEXT_SIBLING_LINENUMBER_KEY));
            buf.append("\"));");
            }
        ***/

        // Make process name, next-lines, and current-lines accessible to
        // gdb:
        /***
            buf.append("gProcessName = pc->p->name();");
            buf.append("gMmdbSB=pc->getScoreboard();");
        ***/
        /***
            buf.append("pcNextLines=pc->getNextMMLineNumbers();");
            buf.append("gAllNextLines=gMmdbSB->getAllNextLines();");
        ***/
        buf.append("gAllCurrentLines=gMmdbSB->getAllCurrentLines();\n");

        /**
         // Only output this info once.
         _nextLineNumbers = null;
        **/

        return buf.toString();
    }

    private String _debugSayClearingLines() {
        if (_DEBUGGING) {
            StringBuffer buf = new StringBuffer();
            buf.append("/*");
            buf.append(DebugUtil.getNodeKind(_node));
            buf.append("*/");
            buf.append("cerr<<pc->p->name()<<\" clearing.\"");
            // String f = _fileName.substring(_fileName.lastIndexOf('/')+1);
            return buf.toString();
            // return new String("cerr<<\"" + f + " clearing.\"");
            // return new String("cerr<<\"" + f + " clearing.\"<<endl;");
        } else {
            return "";
        }
    }

    private String _debugSayAddingLine(String lineNumber) {
        if (_DEBUGGING) {
            String f = _fileName.substring(_fileName.lastIndexOf('/')+1);
            return new String("cerr<<pc->p->name()<<\" adding line "
                    + f
                    + ":"
                    + lineNumber
                    + ".\"<<endl;");
        } else {
            return "";
        }
    }

    private String _debugNewline() {
        if (_DEBUGGING) {
            return new String("<<endl;");
        } else {
            return "";
        }
    }

    private String _debugSayThatsAll() {
        if (_DEBUGGING) {
            String f = _fileName.substring(_fileName.lastIndexOf('/')+1);
            return new String("<<\" @" + f + ":" + _lineNumber + "\"<<endl;");
        } else {
            return "";
        }
    }

    //////////////////////////////////////////////////////////////////////////
    ///         private members

    private static final boolean _DEBUGGING = false;

    private static Object _methodReturnLineNumber = null;

    private Integer  _blockEndLineNumber          = null;
    private boolean  _emittingLineDirectives      = false;
    private String   _fileName                    = null;
    private Integer  _lineNumber                  = null;
    private List     _nextLineNumbers             = null;
    private boolean  _nextLinesAreAddedExplicitly = false;
    private TreeNode _node                        = null;
    private int      _placeholder                 = 0;
    private String   _shortFileName               = null;
}
