/*
  Copyright (c) 2003-2004 The Regents of the University of California.
  All rights reserved.

  Permission is hereby granted, without written agreement and without
  license or royalty fees, to use, copy, modify, and distribute this
  software and its documentation for any purpose, provided that the
  above copyright notice and the following two paragraphs appear in all
  copies of this software and that appropriate acknowledgments are made
  to the research of the Metropolis group.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
  FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
  ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
  THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
  PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
  CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
  ENHANCEMENTS, OR MODIFICATIONS.


  METROPOLIS_COPYRIGHT_VERSION_1
  COPYRIGHTENDKEY

*/
#include "scoreboard.h"
#include "port_rusage.h"
#include "node.h"
#include "programcounter.h"
#include "manager.h"
#include "process.h"
#include "global.h"
#include "statemedium.h"
#include "netlist.h"
#include "medium.h"
#include "action.h"
#include "event.h"
#include "argsutil.h"
#include "nondeterminism.h"
#include <algorithm>
#include "malloc.h"
#include "stdlib.h"
#include "sat_hook.h"

// include behaviors
#include "behavior.h"
#include "DefaultBehavior.h"


scoreboard::scoreboard(sc_module_name name) : sc_channel(name) {
    _pBehavior = new DefaultBehavior();
    //end_module();
}

void scoreboard::registerProcess(ProgramCounter * pc) {
    procList.push_back(pc);
    pc->setScoreboard(this);
}

void scoreboard::registerMedium(medium * m) {
    mediaList.push_back(m);
}

void scoreboard::initialize() {
    std::vector<ProgramCounter *>::iterator procListIter = procList.begin();
    size = procList.size();
    pcVector = new ProgramCounter *[size];
    int i=0;
    for (; procListIter!=procList.end(); procListIter++) {
        pcVector[i] = (*procListIter);
        pcVector[i]->setScoreboard(this);
        i++;
    }
    _mng.pc->setScoreboard(this);

    std::vector<medium *>::iterator mediaListIter = mediaList.begin();
    msize = mediaList.size();
    mediaVector = new medium *[msize];
    i = 0;
    for (; mediaListIter!=mediaList.end(); mediaListIter++) {
        mediaVector[i] = (*mediaListIter);
        i++;
    }

    SatMgr = NULL;

    topNetlist->synch(0);
    if (ltl) topNetlist->ltl();

    //show synchEventGroup
    if (debug_flag) {
        cout<<"**************** synchEventGroup ****************"<<endl;
        cout<<"Total number of synchEvent is "<<synchEventGroup.size()<<endl;
        event * e;
        for (eventGroups::iterator it = synchEventGroup.begin(); it!=synchEventGroup.end(); it++) {
            e = it->first;
            intpair p = it->second;
            cout<<"("<<e->getProcess()->name()<<", ";
            cout<<e->getAction()->getObject()->name()<<", ";
            cout<<e->getAction()->getName()<<") <";
            cout<<p.first<<", "<<p.second<<">"<<endl;
        }
    }
} // end initialize

// process command line arguments
bool scoreboard::processArgs() {
    //default configuration of the simulator
    SchedAlg = &scoreboard::heuristic0;	// heuristic 0
    heuristic = 0;
    interleaving = false;	// use interleaving concurrent simulation
    ltl = false;		// consider any LTL constraints
    ANNOTATION_ENABLED = true; //annotate quantities in simulation
    timestat = false; // no report on time statistics of the simulation
    debug_flag = false; // no debugging info
    debug_level = 0; // level-0 debugging info
    //the larger the debug_level, the more detailed debugging info

    if (getArgc()>1) {
        for (int i=1; i<getArgc(); i++)
            if (!strcmp(getArgv(i), "-d")) {
                debug_flag = true;
                char n;
                ++i;
                if (i==getArgc()) n='E';
                else n=getArgv(i)[0];
                switch(n) {
                case '0':
                    debug_level = 0;
                    break;
                case '1':
                    debug_level = 1;
                    break;
                case '2':
                    debug_level = 2;
                    break;
                case '3':
                    debug_level = 3;
                    break;
                case '4':
                    debug_level = 4;
                    break;
                case '-':
                    --i;
                default:
                    cerr<<"Incorrect argument for -d. Should be '-d <0-4>'."<<endl;
                    cerr<<"Use the default value -d 0."<<endl;
                    debug_level = 0;
                } // end switch
            } else if (!strcmp(getArgv(i), "-ic")) {
                interleaving = true;
            } else if (!strcmp(getArgv(i), "-noqa")) {
                ANNOTATION_ENABLED = false;
            } else if (!strcmp(getArgv(i), "-r")) {
                srandom(getpid());
            } else if (!strcmp(getArgv(i), "-ltl")) {
                ltl = true;
            } else if (!strcmp(getArgv(i), "-t")) {
                timestat = true;
            } else if (!strcmp(getArgv(i), "-h")) {
                char n;
                ++i;
                if (i==getArgc()) n='E';
                else n=getArgv(i)[0];
                switch(n) {
                case '0':
                    SchedAlg = &scoreboard::heuristic0;
                    heuristic = 0;
                    break;
                case '1':
                    SchedAlg = &scoreboard::heuristic1;
                    heuristic = 1;
                    break;
                case '2':
                    SchedAlg = &scoreboard::heuristic2;
                    heuristic = 2;
                    break;
                case '4':
                    SchedAlg = &scoreboard::heuristic2;
                    heuristic = 4;
                    break;
                case '3':
                    SchedAlg = &scoreboard::heuristic3;
                    heuristic = 3;
                    break;
                case '-':
                    --i;
                default:
                    cerr<<"Incorrect argument for -h. Should be '-h <0-3>'."<<endl;
                    cerr<<"Use the default heuristic 0."<<endl;
                    SchedAlg = &scoreboard::heuristic0;
                } // end switch
            } else if (!strcmp(getArgv(i), "-help")) {
                cout<<endl;
                cout<<getArgv(0)<<"  [-help]  [-d <0-4>]  [-r]  [-t]  [-h <0-3>]  [-ltl]  [-ic] [-noqa]"<<endl;
                cout<<"\t-help\t\tShow this message"<<endl;
                cout<<"\t-d <0-4>\tShow scoreboard for each step during simulation."<<endl;
                cout<<"\t\t\tA larger number gives more detailed information."<<endl;
                cout<<"\t-r\t\tExhibit true (not pseudo) randomness  during simulation"<<endl;
                cout<<"\t-t\t\tReport time statistics on simulation"<<endl;
                cout<<"\t-h <0-4>\tChoose different scheduling heuristics"<<endl;
                cout<<"\t    0\t\tThe default scheduling heuristic (the original one)"<<endl;
                cout<<"\t    1\t\tRandomly choose one runnable process"<<endl;
                cout<<"\t    2\t\tChoose a (sub)set of compatible and runnable processes"<<endl;
                cout<<"\t    3\t\tChoose a maxium set of compatible and runnable processes"<<endl;
                cout<<"\t    4\t\tChoose heuristic 2 with hook function to guide SAT solver"<<endl;
                cout<<"\t-ltl\t\tEnforce LTL constraints during simulation"<<endl;
	        cout<<"\t-ic\t\tUse interleaving concurrent feature in simulation"<<endl;
	        cout<<"\t-noqa\t\tDo not annotate quantities during simulation"<<endl;
                return true;
            } //end if
    }// end for
    return false;
} // end processArg

//Find out if there are any await statements waiting to evaulate conditions.
bool scoreboard::eval() {
    bool flag = false;
    for (int i=0; i<size; i++) {
	if (pcVector[i]->SchedState == SchedStateVal::EVALUATE) {
	    notify(DeltaCycle, pcVector[i]->invoke);
	    flag = true;
	    wait(_mng.pc->invoke);
	}
    }
    return flag;
}

//Before restart of manager cycle, change DONTRUN to UNKNOW for
//interface function, block or label
//For await statements, create csOrder[].
void scoreboard::preSet() {
    testSchedFlag = false;

    for (int i=0; i<size; i++) {
        pcVector[i]->savePendingEvents();
        pcVector[i]->_pcBlocked.clear();

        if (pcVector[i]->SchedState == SchedStateVal::TESTSCHEDULING)
            testSchedFlag = true;

        if (pcVector[i]->SchedState == SchedStateVal::DONTRUN)
            pcVector[i]->SchedState = SchedStateVal::UNKNOW;
    } // end for

    if (numSynchEventGroup>0) {
        collect_ref_count();
        process_equalvars_in_synch();
    }

    for (int i=0; i<numSynchEventGroup; i++) {
        if (enabledSynchEventEachGroup[i] < numSynchEventEachGroup[i]) {
            for (int j=0; j<enabledSynchEventEachGroup[i]; j++) {
                enabledSynchEventPC[i][j]->freeAllPendingEvents();
                enabledSynchEventPC[i][j]->SchedState = SchedStateVal::DONTRUN;
                enabledSynchEventPC[i][j] = NULL;
            } // end for
            enabledSynchEventEachGroup[i] = 0;
            //} else {
            //cout<<"One group of synch events are ready to go all together."<<endl;
        } // end if
    } // end for
} // end preSet

//Adjust pc->selected to actual choice stored in csOrder[]
//Clean up pc->csOrder[] created in preSet()
void scoreboard::postSet() {
    for (int i = 0; i < size; i++) {
        ProgramCounter *pc = pcVector[i];

        if (pc->SchedState == SchedStateVal::FINISHED) continue;

        if (((!testSchedFlag) &&
                    ((pc->SchedState & FunctionType::ANNOTATION) == 0) &&
                    (!pc->_inSynch) &&
                    (pc->SchedState == SchedStateVal::DONTRUN)) //||
                //(testSchedFlag && pc->mode==RunType::EXEC)
            ) {
            pc->loadPendingEvents();
        }
    } // end for

    for (int i=0; i<numSynchEventGroup; i++) {
        enabledSynchEventEachGroup[i] = 0;
    } // end for
} // end postSet

/*
 * pass1
 * Check first deterministic processes (the one with 1 pending event).
 * Update the reference count value in the eventSynchMap_t to order the
 * critical section selection next.
 * Then checks each process with all the others in order to detect any
 * conflict in test list.
 */
bool scoreboard::pass1() {
    bool rtn = false;

    for (eventSynchMap_t::iterator it = _eventSynchMap.begin();
	 it != _eventSynchMap.end();
	 it++) {
        it->first->_refCount = 0;
        it->second->_refCount = 0;
    } // end for

    // nPendingEvents == 1 --> check deterministic critical sections selection
    // nPendingEvents == 2 --> check non deterministic critical sections selection
    for (int nPendingEvents = 1; nPendingEvents <= 2; nPendingEvents++) {
        for (int i = 0; i < size; i++) {
            ProgramCounter * pc = pcVector[i];

            if (testSchedFlag) {
                if (pc->SchedState != SchedStateVal::TESTSCHEDULING) continue;
            } else {
                if (pc->SchedState != SchedStateVal::UNKNOW) continue;
            } // end if

            if (pc->_pendingEvents.size() == 0) continue;

            if (nPendingEvents == 1) {
                if (pc->_pendingEvents.size() == 1) resolveSynch0(pc);		// deterministic => _pendingEvents.size() == 1
            } else {
                if (pc->_pendingEvents.size() > 1) resolveSynch1(pc);		// non deterministic => _pendingEvents.size() > 1

                // suppose all label or block could run
                if (pc->funcType & FunctionType::LABEL) {
                    rtn = true;
                    continue;
                } // end if

                // medium-centric approach
                rtn = couldRun(pc);

                /* process-centric approach
                   int j;

                   do {
                   //					bool first = true;
                   for (j = 1; j < size; j++) {
                   if (i == j) continue;
                   ProgramCounter * cmp = pcVector[j];

                   if (cmp->SchedState != SchedStateVal::UNKNOW) continue;
                   if (cmp->_pendingEvents.empty()) continue;

                   if (pc->funcType & FunctionType::AWAIT) {
                   if (!couldRunAwait(pc, cmp)) {
                   break;
                   } // end if
                   } else if (pc->funcType & FunctionType::INTFCFUNC) {
                   if (!couldRunIntfc(pc, cmp)) {
                   break;
                   } // end if
                   } // end if
                   } // end for
                   } while ((j < size) && (pc->_pendingEvents.size() > 0));
                   if (j == size) rtn = true;
                */
            } // end if
        } // end for
    } // end for

    return (rtn || testSchedFlag);
}



// Second scan. Find out wether there is any inconsistence if all processes from pass1 begin to run.
bool scoreboard::pass2() {
    bool bFound = false;

    // mode == 0 => collect blocking process relations
    // mode == 1 => perform necessary updates (setMustDo / setMustNotDo)
    for (int mode = 0; mode < 2; mode++) {
        for (int k=0; k<msize; k++) { // iterate over all media
            medium * m = mediaVector[k];
            for (potentialUserPCs_t::iterator i = m->_potentialUserPCs.begin(); i != m->_potentialUserPCs.end(); i++) {
                ProgramCounter *pc = (*i);

                if (testSchedFlag) {
                    if (pc->SchedState != SchedStateVal::TESTSCHEDULING) continue;
                } else {
                    if (pc->SchedState != SchedStateVal::UNKNOW) continue;
                } // end if

                if (pc->_pendingEvents.empty()) continue;

                bool runBothFlag = true;
                potentialUserPCs_t::iterator j = i;
                for (j++; j != m->_potentialUserPCs.end(); j++) {
                    ProgramCounter *cmp = (*j);

                    if (testSchedFlag) {
                        if (cmp->SchedState != SchedStateVal::TESTSCHEDULING) continue;
                    } else {
                        if (cmp->SchedState != SchedStateVal::UNKNOW) continue;
                    } // end if

                    if (cmp->_pendingEvents.empty()) continue;

                    if (!couldRunBoth(mode, pc, cmp)) {
                        runBothFlag = false;
                        if (mode == 0) {	// collect information on blocking process relations
                            set2blocked(pc, cmp);
                            set2blocked(cmp, pc);
                        } else				// perform event/process selection according to user specific simulator plugin
                            _pBehavior->resolve(pc, cmp);
                    } // end if
                } // end for j
                if (runBothFlag) bFound = true;
            } // end for i
        } // end for k
    } // end for mode

    return (bFound || testSchedFlag);
} // end pass2

bool scoreboard::resolveSynch0(ProgramCounter *pPC) {
    // prepare reference count number
    _maxRefCount = 0;
    // prepare critical section selection order
    pPC->_csOrder.clear();

    // pPC->_pendingEvents.size() == 1
    for (eventMap_t::iterator it = pPC->_pendingEvents.begin(); it != pPC->_pendingEvents.end(); it++) {
        event *pEvent = (*it).first;	// event
        short nCs = (*it).second;		// critical section number

        pPC->_csOrder.push_back(nCs);

        // get the corresponding synch event in the map
        EventMap *pFoundEM = getEventSynchMap(pEvent, true);

        if (pFoundEM) {
            // count how many times this event is pending in all non deterministic processes (pending events > 1)
            for (int i = 0; i < size; i++) {
                ProgramCounter *pc = pcVector[i];


                if (pc->SchedState != SchedStateVal::UNKNOW) continue;
                if (pc->_pendingEvents.empty()) continue;

                for (eventMap_t::iterator it = pc->_pendingEvents.begin(); it != pc->_pendingEvents.end(); it++) {
                    if (it->first == pFoundEM->_e && pc->_pendingEvents.size() > 1) {

                        pFoundEM->_refCount++;
                        if (pFoundEM->_refCount > _maxRefCount) _maxRefCount = pFoundEM->_refCount;
                        break;
                    } // end if
                } // end for
            } // end for
        } // end if
    } // end for

    return true;

} // end resolveSynch0

bool scoreboard::resolveSynch1(ProgramCounter *pPC) {
    // prepare critical section selection order
    pPC->_csOrder.clear();

    // start considering pending events corresponding to the highest
    // value of the reference count
    for (int i = _maxRefCount; i >= 0; --i) {
        // store first position of the group of critical sections
        // having the same reference count in order to randomize the
        // order among the critical sections with the same reference count
        // value
        int posInsCsOrder = pPC->_csOrder.size();
        for (eventMap_t::iterator it = pPC->_pendingEvents.begin(); it != pPC->_pendingEvents.end(); it++) {
            event *pEvent = it->first;	// event
            short nCs = it->second;		// critical section number

            // get pEvent position in the event synch map
            EventMap *pFoundEM = getEventSynchMap(pEvent, false);

            bool bInsertCS = false;

            if (pFoundEM) {
                if (pFoundEM->_refCount == i) bInsertCS = true;
            } else
                bInsertCS = true;

            if (bInsertCS) {
                // check if the critical section has already been inserted into _csOrder
                intVect_t::iterator it = std::find(pPC->_csOrder.begin(), pPC->_csOrder.end(), nCs);
                if (it == pPC->_csOrder.end()) {
                    // randomize critical section order
                    int dim = pPC->_csOrder.size();

                    int pos = (dim - posInsCsOrder == 0) ? 0 : (random() % (dim - posInsCsOrder));
                    if (posInsCsOrder + pos >= dim)
                        pPC->_csOrder.push_back(nCs);
                    else {
                        short oldCs = pPC->_csOrder[posInsCsOrder + pos];
                        pPC->_csOrder[posInsCsOrder + pos] = nCs;
                        pPC->_csOrder.push_back(oldCs);
                    } // end if
                } // end if

            } // end if
        } // end if
    } // end for

    return true;
} // end resolveSynch1

//Third scan. Constraints check. Not yet implemented.
bool scoreboard::pass3() {

    return true;
}

//Forth scan. Do user specified _doScheduling.
//bool pass4(){} = _doScheduling in Scheduler

//A little like pass2(). Collect all conflicts among all pending events
void scoreboard::collectEventConflicts() {
    conflictRSet.clear();
    eventVarIndex.clear();
    eventMap_t  pcBackup, cmpBackup;

    for (int i = 0; i < size; i++) {
        ProgramCounter *pc = pcVector[i];

        if (testSchedFlag) {
            if (pc->SchedState != SchedStateVal::TESTSCHEDULING) continue;
        } else {
            if (pc->SchedState != SchedStateVal::UNKNOW) continue;
        } // end if

        if (pc->_pendingEvents.empty()) continue;

        for (eventMap_t::iterator it=pc->_pendingEvents.begin(); it!=pc->_pendingEvents.end(); it++) {
            eventVarIndex[it->first] = 0;
            if (pc->funcType & FunctionType::AWAIT) {
                ++it;
                eventMap_t::iterator it2=it;
                --it;
                for (; it2!=pc->_pendingEvents.end(); it2++)
                    conflictRSet.insert(std::make_pair(it->first, it2->first));
            } // end if
        } // end for

        if (pc->funcType & FunctionType::LABEL) continue;

        pcBackup   = pc->_pendingEvents;
        for (int j = i + 1; j < size; j++) {
            ProgramCounter *cmp = pcVector[j];

            if (testSchedFlag) {
                if (cmp->SchedState != SchedStateVal::TESTSCHEDULING) continue;
            } else {
                if (cmp->SchedState != SchedStateVal::UNKNOW) continue;
            } // end if

            if (cmp->_pendingEvents.empty()) continue;

            if (cmp->funcType & FunctionType::LABEL) continue;

            if ((pc->funcType & FunctionType::INTFCFUNC) &&
                    (cmp->funcType & FunctionType::INTFCFUNC)) continue;

            cmpBackup= cmp->_pendingEvents;

            for (eventMap_t::iterator itpc=pcBackup.begin(); itpc!=pcBackup.end(); itpc++)
                for (eventMap_t::iterator itcmp=cmpBackup.begin(); itcmp!=cmpBackup.end(); itcmp++) {
                    pc->_pendingEvents.clear();
                    cmp->_pendingEvents.clear();
                    pc->_pendingEvents.insert(*itpc);
                    cmp->_pendingEvents.insert(*itcmp);
                    if (!couldRunBoth(0, pc, cmp))
                        conflictRSet.insert(std::make_pair(itpc->first, itcmp->first));
                } // end for

            cmp->_pendingEvents = cmpBackup;
        } // end for j
        pc->_pendingEvents = pcBackup;
    } // end for i
} // end collectEventConflicts


//self-check done by individual process
// if ANNOTATION exists, it won't get here.
void scoreboard::check(ProgramCounter *pc) {
    /*
      if (!interleaving) {
      // original non-interleaving concurrent version
      if (pc->SchedState == SchedStateVal::UNKNOW)
      pc->SchedState = SchedStateVal::DONTRUN;
      else // pc->SchedState == SchedStateVal::TESTSCHEDULING
      pc->SchedState = SchedStateVal::RUN;
      } else
    */
    {
        // interleaving concurrent version
        switch (pc->funcType) {
        case FunctionType::LABEL:
            if (getEventSynchMap( pc->_pendingEvents.begin()->first, false)) {
                pc->SchedState =SchedStateVal::DONTRUN;
            } else {
                pc->SchedState =SchedStateVal::RUN;
            } // end if
            break;
        case FunctionType::INTFCFUNC:
            if (getEventSynchMap( pc->_pendingEvents.begin()->first, false)) {
                pc->SchedState =SchedStateVal::DONTRUN;
            } else {
                for (int j = 1; j < size; j++) {
                    if (pc == pcVector[j]) continue;
                    if (!couldRunIntfc(pc, pcVector[j])) {
                        pc->SchedState =SchedStateVal::DONTRUN;
                        return;
                    } // end if
                } // end for

                pc->SchedState =SchedStateVal::RUN;
            } // end if
            break;
        case FunctionType::AWAIT:
            //favor synched CS over non-synched CS by defering
            // the scheduling to (*SchedAlg)()
            for (eventMap_t::iterator it = pc->_pendingEvents.begin();
                 it != pc->_pendingEvents.end();
                 it++) {
                if (getEventSynchMap( it->first, false)) {
                    pc->SchedState =SchedStateVal::DONTRUN;
                    return;
                } // end if
            } //end for

            for (int j = 1; j < size; j++) {
                if (pc == pcVector[j]) continue;
                if (!couldRunAwait(pc, pcVector[j])) {
                    pc->SchedState =SchedStateVal::DONTRUN;
                    return;
                } // end if
            } // end for

            if (pc->_pendingEvents.size() == 1) {
                // only 1 pending event => RUN the corresponding process
                pc->selected = pc->_pendingEvents.begin()->second;
            } else if (pc->_pendingEvents.size() > 1) {
                // more than 1 pending event => choose a random critical
                // section and RUN it
                int choice = (int)(random() % (pc->_pendingEvents.size()));
                eventMap_t::iterator it = pc->_pendingEvents.begin();
                for (int i = 0 ;
                     it != pc->_pendingEvents.end() && i < choice;
                     it++, i++);
                pc->selected = it->second;
            } // end if
            pc->SchedState =SchedStateVal::RUN;

            break;
        case FunctionType::ANNOTATION: // should never happen
        default:
            cerr<<"Internal Error! Exit."<<endl;
            sc_stop();
        } // end switch
    } // end if
} // end check

//Update corresponding synch event info
void scoreboard::updateSynchEventInfo(int eventGroupID, ProgramCounter* eventPC) {
    enabledSynchEventPC[eventGroupID][enabledSynchEventEachGroup[eventGroupID]] = eventPC;
    enabledSynchEventEachGroup[eventGroupID]++;
    eventPC->_inSynch = true;
} // end updateSynchEventInfo

//Return whether all events in the ith synch event group are enabled
bool scoreboard::isSynchEventGroupEnabled(int ith) {
    return (enabledSynchEventEachGroup[ith]==numSynchEventEachGroup[ith]);
} // end isSynchEventGroupEnabled

//Due to the equal variables comparison, enable/disable this group
//of synched events
void scoreboard::setSynchEventGroupStatus(int i, bool satisfied) {
    if (!satisfied) {
        for (int j=0; j<enabledSynchEventEachGroup[i]; j++) {
            enabledSynchEventPC[i][j]->freeAllPendingEvents();
            enabledSynchEventPC[i][j]->SchedState = SchedStateVal::DONTRUN;
            enabledSynchEventPC[i][j] = NULL;
        } // end for
        enabledSynchEventEachGroup[i] = 0;
    } // end if
} // end setSynchEventGroupStatus

//add (synch event, <beg group, end group>) info to scoreboard
void scoreboard::addSynchEventGroupInfo(process *p, medium *m, const char *c, enum ActionState as, int groupNum) {
    event * key = getEvent(p, m, c, ACTION_STATE_BEGIN);
    int beg, end;
    beg = end = -1;
    if (as == ACTION_STATE_BEGIN)
        beg = groupNum;
    else
        end = groupNum;
    eventGroups::iterator it = synchEventGroup.find(key);
    if (it == synchEventGroup.end()) {
        bool inserted;
        pair<eventGroups::iterator, bool>(it, inserted) = synchEventGroup.insert(eventGroups::value_type(key, intpair(beg, end)));
    } else {
        if (as == ACTION_STATE_BEGIN)
            it->second.first = groupNum;
        else
            it->second.second = groupNum;
    }
} // end addSynchEventGroupInfo

//Collect the reference count of a CS in await
void scoreboard::collect_ref_count() {
    int beg, end; //total # synch events - # enabled synch events
    for (int i=0; i<size; i++) {
        ProgramCounter * pc = pcVector[i];
        if ((pc->funcType & FunctionType::AWAIT) != 0) {
            if (pc->_pendingEvents.size()<=1) continue;
            eventMap_t::iterator it = pc->_pendingEvents.begin();
            for (; it!=pc->_pendingEvents.end(); it++) {
                eventGroups::iterator eit = synchEventGroup.find(it->first);
                if (eit == synchEventGroup.end()) {
                    it->first->setRefCount(INT_MAX);
                    continue;
                }
                intpair p = eit->second;
                if (p.first >= 0)
                    beg = numSynchEventEachGroup[p.first]-enabledSynchEventEachGroup[p.first];
                else
                    beg = INT_MAX;

                if (p.second >= 0)
                    end = numSynchEventEachGroup[p.second]-enabledSynchEventEachGroup[p.second];
                else
                    end = INT_MAX;

                it->first->setRefCount((beg<end) ? beg : end);
            } // end for it
        } // end if pc->funcType
    } // end for i
} // end collect_ref_count

bool scoreboard::resolveSynch() {
    for (int i=0; i<numSynchEventGroup; i++) {
        if (numSynchEventEachGroup[i] == enabledSynchEventEachGroup[i]) {
            bool enabled = true;
            for (int j=0; (j<enabledSynchEventEachGroup[i]) & enabled; j++) {
                if (enabledSynchEventPC[i][j]->_pendingEvents.size()==0) enabled = false;
            } // end for
            if (!enabled) {
                for (int j=0; j<enabledSynchEventEachGroup[i]; j++) {
                    enabledSynchEventPC[i][j]->freeAllPendingEvents();
                    enabledSynchEventPC[i][j]->SchedState = SchedStateVal::DONTRUN;
                    enabledSynchEventPC[i][j] = NULL;
                } // end for
                enabledSynchEventEachGroup[i] = 0;
            } //end if
        } // end if
    } // end for

    return true;
    //return topNetlist->resolveSynch();
} // end resolveSynch

/*
 * set2run
 * set the schedule state of processes to RUN according to the rules specified
 * in the function
 h */
bool scoreboard::set2run() {
    bool bContinue = false;

    for (int i = 0; i < size; i++) {
        ProgramCounter *pc = pcVector[i];
        if (pc->SchedState == SchedStateVal::RUN) bContinue = true;

        // if test scheduling and the process is in test scheduling
        // or
        // if not test scheduling and the process has UNKNOW schedule state
        if ((testSchedFlag && pc->SchedState == SchedStateVal::TESTSCHEDULING) ||
                (!testSchedFlag && pc->SchedState == SchedStateVal::UNKNOW)) {
            if (pc->_pendingEvents.size() == 0) {			// no pending events => DONTRUN
                pc->SchedState = SchedStateVal::DONTRUN;
            } else if (pc->_pendingEvents.size() == 1) {	// only 1 pending event => RUN the corresponding process
                pc->SchedState = SchedStateVal::RUN;
                bContinue = true;
                pc->selected = pc->_pendingEvents.begin()->second;
            } else if (pc->_pendingEvents.size() > 1) {		// more than 1 pending event => choose a random critical section and RUN it
                std::vector<event*> tmp;
                std::vector<int> tmpsel;
                std::vector<event*> orig;
                std::vector<int> origsel;
                eventMap_t::iterator it = pc->_pendingEvents.begin();
                int min = INT_MAX;
                int refc;
                for (; it != pc->_pendingEvents.end(); it++) {
                    refc = it->first->getRefCount();
                    orig.push_back(it->first);
                    origsel.push_back(it->second);
                    if (refc > min) continue;
                    if (refc < min) {
                        tmp.clear();
                        tmpsel.clear();
                        min = refc;
                    }
                    tmp.push_back(it->first);
                    tmpsel.push_back(it->second);
                }

                pc->SchedState = SchedStateVal::RUN;
                bContinue = true;
                int choice;
                event* selectedEvent;
                if (tmp.size() > 0) {
                    choice = (int)(random() % (tmp.size()));
                    selectedEvent = tmp[choice];
                    pc->selected = tmpsel[choice];
                } else {
                    choice = (int)(random() % (pc->_pendingEvents.size()));
                    selectedEvent = orig[choice];
                    pc->selected = origsel[choice];
                }
                pc->freeAllPendingEvents();
                pc->addPendingEvent(selectedEvent, pc->selected);
            } // end if
        } //else
        //pc->SchedState = SchedStateVal::DONTRUN;
    } // end for

    return (bContinue || testSchedFlag);
} // end set2run

/*
 * set2dontrun
 * set the schedule state of processes to DONTRUN performing some necessary cleanup steps
 */
void scoreboard::set2dontrun(ProgramCounter * pc) {
    // look if this process is in synch with another one => set to DONTRUN the corresponding process
    for (eventMap_t::iterator it = pc->_pendingEvents.begin(); it != pc->_pendingEvents.end(); it++) {
        event *pEvent = it->first;

        for (eventSynchMap_t::iterator it_map = _eventSynchMap.begin(); it_map != _eventSynchMap.end(); it_map++) {
            EventMap *pEventMap = NULL;
            if (pEvent == it_map->first->_e) pEventMap = it_map->second;
            if (pEvent == it_map->second->_e) pEventMap = it_map->first;
            if (pEventMap) {
                if (pEventMap->_refCount != -1) {
                    pEventMap->_refCount = -1;
                    set2dontrun(pEventMap->_e->getProcess()->pc);
                }
            }
        } // end for

    } // end for

    // delete any pending event
    pc->_pendingEvents.clear();
    // set to DONTRUN
    pc->SchedState = SchedStateVal::DONTRUN;
} // end set2dontrun

/*
 * set2blocked
 * fill the vector of _pcBlocked according to synchs
 */
void scoreboard::set2blocked(ProgramCounter * pc, ProgramCounter * pc_blocked) {
    // search if the pc_blocked is already present in the list
    pcList_t::iterator it = std::find(pc->_pcBlocked.begin(), pc->_pcBlocked.end(), pc_blocked);
    if (it != pc->_pcBlocked.end()) return;

    // add the blocked pc "pc_blocked" to pc
    pc->_pcBlocked.push_back(pc_blocked);

    // fill the blocked queue according to synchs
    for (eventMap_t::iterator it = pc_blocked->_pendingEvents.begin(); it != pc_blocked->_pendingEvents.end(); it++) {
        event *pEvent = it->first;

        for (eventSynchMap_t::iterator it_map = _eventSynchMap.begin(); it_map != _eventSynchMap.end(); it_map++) {
            EventMap *pEventMap = NULL;
            if (pEvent == it_map->first->_e) pEventMap = it_map->second;
            if (pEvent == it_map->second->_e) pEventMap = it_map->first;
            if (pEventMap)
                set2blocked(pc, pEventMap->_e->getProcess()->pc);

        } // end for
    } // end for
} // end set2blocked

//Could pc run?
bool scoreboard::couldRun(ProgramCounter * pc) {
    if (pc->funcType == FunctionType::AWAIT) {
        for (eventMap_t::iterator it = pc->_pendingEvents.begin(); it != pc->_pendingEvents.end(); ) {

            bool flag = false;

            _2DList_t *pSetListAll_pc = pc->setList.back();
            simpleList_t *pSetList_pc = (*pSetListAll_pc)[it->second];

            //check if pc has setList intersecting with existing prevented interface list
            for (int i = 0; i < pSetList_pc->size(); i++) {
                node * n = (*pSetList_pc)[i];
                flag = ((medium*)n->object)->isIntfcPrevented((n->intfcName).c_str(), pc);
                if (flag) break;
            } // end for

            //check if pc has testList intersecting with interfaces being used
            if (!flag) {
                simpleList_t::iterator it_nl;
                intList_t::iterator it_ssl;

                _2DList_t *pTestListAll_pc = pc->testList.back();
                simpleList_t *pTestList_pc = (*pTestListAll_pc)[it->second];
                for (int i = 0; i < pTestList_pc->size(); i++) {
                    node * n = (*pTestList_pc)[i];
                    flag = ((medium*)n->object)->isIntfcUsed((n->intfcName).c_str(), pc);
                    if (flag) break;
                } // end for
            } //end if

            if (flag) {
                ++it;
                if (it != pc->_pendingEvents.end()) {
                    event * pKey = it->first;
                    pc->_pendingEvents.erase(--it);
                    it = pc->_pendingEvents.find(pKey);
                } else {
                    pc->_pendingEvents.erase(--it);
                    it = pc->_pendingEvents.end();
                }// end if
            } else
                it++;
        } // end for

        return (pc->_pendingEvents.size() > 0);
    } else if (pc->funcType == FunctionType::INTFCFUNC) {
        bool flag = false;

        //check if interface is being prevented
        flag = ((medium*)pc->obj.object)->isIntfcPrevented(pc->obj.intfcName.c_str(), pc);

        if (flag) pc->_pendingEvents.clear();

        return (!flag);
    } else {
        return true;
    } // end if

} // end couldRun

//Could pc(await statement) run under the influence of cmp(DONTRUN/UNKNOW/RUNNING)?
bool scoreboard::couldRunAwait(ProgramCounter * pc, ProgramCounter * cmp) {
    for (eventMap_t::iterator it = pc->_pendingEvents.begin(); it != pc->_pendingEvents.end(); ) {

        bool flag = false;

        _2DList_t *pSetListAll_pc = pc->setList.back();
        simpleList_t *pSetList_pc = (*pSetListAll_pc)[it->second];

        //check if pc has a setList intersecting with a cmp preventList
        for (int i = 0; i < cmp->preventList.size(); i++) {
            simpleList_t *pSetList_cmp = cmp->preventList[i];
            //			flag = contained(&(pc->obj), pSetList_cmp);
            if (!flag) flag = intersect(pSetList_cmp, pSetList_pc);
            if (flag) break;
        } // end for

        //		//check if cmp has any running program that pc is waiting for finishing
        //		if (flag) {
        //			it++;
        //			continue;
        //		} // end if

        if (!flag) {
            simpleList_t::iterator it_nl;
            intList_t::iterator it_ssl;

            _2DList_t *pTestListAll_pc = pc->testList.back();
            simpleList_t *pTestList_pc = (*pTestListAll_pc)[it->second];
            for (it_nl = cmp->nodeList.begin(), it_ssl = cmp->SchedStateList.begin(); it_nl != cmp->nodeList.end() && it_ssl != cmp->SchedStateList.end(); it_nl++, it_ssl++) {
                int state = *(*it_ssl);
                if (state != SchedStateVal::RUNNING) continue;
                flag = contained(*it_nl, pTestList_pc);
                if (flag) break;
            } // end for
        } //end if

        if (flag) {
            ++it;
            if (it != pc->_pendingEvents.end()) {
                event * pKey = it->first;
                pc->_pendingEvents.erase(--it);
                it = pc->_pendingEvents.find(pKey);
            } else {
                pc->_pendingEvents.erase(--it);
                it = pc->_pendingEvents.end();
            }// end if
        } else
            it++;
    } // end for

    return (pc->_pendingEvents.size() > 0);
}

//Could pc(interface function) run under the influence of cmp(DONTRUN/UNKNOW/RUNNING)?
bool scoreboard::couldRunIntfc(ProgramCounter * pc, ProgramCounter * cmp) {
    bool flag = false;

    //check if cmp(->preventList) prevents pc(->node) from running
    for (int j = 0; j < cmp->preventList.size(); j++) {
        simpleList_t *pPreventList = cmp->preventList[j];
        flag = contained(&(pc->obj), pPreventList);
        if (flag) break;
    } // end for

    if (flag) pc->_pendingEvents.clear();

    return (!flag);
}

//Could pc run together with cmp
bool scoreboard::couldRunBoth(int mode, ProgramCounter * pc, ProgramCounter * cmp) {

    event *pEvent_pc = NULL, *pEvent_cmp = NULL;

    for (intVect_t::iterator it_cs_pc = pc->_csOrder.begin(); (!pEvent_pc && !pEvent_cmp) && it_cs_pc != pc->_csOrder.end(); it_cs_pc++) {
        for (intVect_t::iterator it_cs_cmp = cmp->_csOrder.begin(); (!pEvent_pc && !pEvent_cmp) && it_cs_cmp != cmp->_csOrder.end(); it_cs_cmp++) {
            if (cmp->funcType & FunctionType::AWAIT) {
                _2DList_t *pSetListAll_cmp = cmp->setList.back();
                _2DList_t *pTestListAll_cmp = cmp->testList.back();

                simpleList_t *pSetList_cmp = (*pSetListAll_cmp)[*it_cs_cmp];
                simpleList_t *pTestList_cmp = (*pTestListAll_cmp)[*it_cs_cmp];
                if (contained(&(pc->obj), pSetList_cmp)) continue;

                if (pc->funcType & FunctionType::AWAIT) {
                    _2DList_t *pSetListAll_pc = pc->setList.back();
                    _2DList_t *pTestListAll_pc = pc->testList.back();

                    simpleList_t *pSetList_pc = (*pSetListAll_pc)[*it_cs_pc];
                    simpleList_t *pTestList_pc = (*pTestListAll_pc)[*it_cs_pc];

                    if (contained(&(cmp->obj), pSetList_pc)) continue;

                    //check if cmp(->setList[selected]) and pc(->setList[selected]) intersect
                    //if (intersect(_SetNodes1, _SetNodes2)) return false;


                    //check if cmp(->setList[selected]) and pc(->testList[selected]) intersect
                    if (intersect(pSetList_cmp, pTestList_pc)) continue;

                    //check if cmp(->testList[selected]) and pc(->setList[selected]) intersect
                    if (intersect(pTestList_cmp, pSetList_pc)) continue;
                } // end if
            } else {
                if (pc->funcType & FunctionType::AWAIT) {
                    _2DList_t *pSetListAll_pc = pc->setList.back();
                    _2DList_t *pTestListAll_pc = pc->testList.back();

                    simpleList_t *pSetList_pc = (*pSetListAll_pc)[*it_cs_pc];
                    simpleList_t *pTestList_pc = (*pTestListAll_pc)[*it_cs_pc];

                    if (contained(&(cmp->obj), pSetList_pc)) continue;
                } // end if
            } // end if

            // search the events in the pending event list of the corresponding processes
            if (pc->funcType & FunctionType::AWAIT) {
                for (eventMap_t::iterator it = pc->_pendingEvents.begin(); !pEvent_pc && it != pc->_pendingEvents.end(); it++) {
                    if (it->second == *it_cs_pc) pEvent_pc = it->first;
                } // end for
            } else {
                if (pc->_pendingEvents.size() == 1) pEvent_pc = pc->_pendingEvents.begin()->first;
            } // end if

            if (cmp->funcType & FunctionType::AWAIT) {
                for (eventMap_t::iterator it = cmp->_pendingEvents.begin(); !pEvent_cmp && it != cmp->_pendingEvents.end(); it++) {
                    if (it->second == *it_cs_cmp) pEvent_cmp = it->first;
                } // end for
            } else {
                if (cmp->_pendingEvents.size() == 1) pEvent_cmp = cmp->_pendingEvents.begin()->first;
            } // end if

            // in case both processes pc and cmp are nondeterministic, check if the critical sections
            // of the selected events prevent them to be executed according to the synch map

            if (pEvent_pc && pEvent_cmp) {
                // get corresponding synch event for pc
                EventMap *pEvent_pc_map = getEventSynchMap(pEvent_pc, true);
                // get corresponding synch event for cmp
                EventMap *pEvent_cmp_map = getEventSynchMap(pEvent_cmp, true);

                if (((pEvent_pc_map) &&
                            // found and cmp contains the requested event (pEvent_pc_map->_e) but it has not been selected
                            // (pEvent_cmp != pEvent_pc_map->_e)
                            (cmp->containsPendingEvent(pEvent_pc_map->_e) && pEvent_cmp != pEvent_pc_map->_e)) ||
                        ((pEvent_cmp_map) &&
                                // found and pc contains the requested event (pEvent_cmp_map->_e) but it has not been selected
                                // (pEvent_pc != pEvent_cmp_map->_e)
                                (pc->containsPendingEvent(pEvent_cmp_map->_e) && pEvent_pc != pEvent_cmp_map->_e))) {
                    pEvent_pc = NULL;	// abort

                    pEvent_cmp = NULL;	// abort
                } // end if
            } // end if
        } // end for
    } // end for

    // update
    if (mode == 1) {
        if (pEvent_pc && pEvent_cmp) {
            pc->setMustDo(pEvent_pc);
            cmp->setMustDo(pEvent_cmp);
        } // end if
    } // end if

    return (pEvent_pc && pEvent_cmp);
} // end couldRunBoth

//check if one node is in a set of nodes
bool scoreboard::contained(node * n, simpleList_t *pList) {

    bool rtn = false;
    for (int i = 0; i < pList->size(); i++) {
        node *pNode = (*pList)[i];
        if (n->object != pNode->object) continue;
        if ((pNode->intfcName != "all") && (n->intfcName != "all") && (pNode->intfcName != n->intfcName) ) continue;
        else {
            rtn = true;
            break;
        } // end if
    } // end for
    return rtn;
} // end contained

//check if two sets of nodes intersects
bool scoreboard::intersect(simpleList_t * pList1, simpleList_t * pList2) {
    bool rtn = false;
    for (int i = 0; i < pList1->size(); i++) {
        node *pNode = (*pList1)[i];
        if (contained(pNode, pList2)) {
            rtn = true;
            break;
        } // end if
    } // end for
    return rtn;
} // end intersect

//print out all scoreboard

void scoreboard::showSB() {
    ProgramCounter * pc;

    String state[8]={"", "RUN", "DONTRUN", "RUNNING", "UNKNOW", "EVALUATE", "FINISHED", "TESTSCHEDULING"};
    String ftype[5]={"", "AWAIT", "LABEL", "INTFCFUNC", "ANNOTATION"};
    String astate[2]={"BEGIN", "END"};
    String runMode[2]={"EXEC", "TRY"};

    cout<<"************************Scoreboard************************"<<endl;
    for (int i = 0; i < size; i++) {

        ProgramCounter *pc = pcVector[i];

        cout<<"ProgramCounter #"<<i+1<<"\t"<<pc<<endl;
        cout<<"process * p="<<(pc->p)<<endl;
        cout<<"process name="<<pc->p->name()<<endl;
        cout<<"process id="<<pc->p->_id<<endl;
        cout<<"int mode="<<runMode[pc->mode]<<endl;
        cout<<"node obj={"<<pc->obj.object->name()<<", "<<pc->obj.intfcName.c_str()<<"}"<<endl;
        cout<<"String stLabel="<<pc->stLabel.c_str()<<endl;
        cout<<"int SchedState="<<state[pc->SchedState]<<endl;

        cout<<"pending events={"<<endl;
        for (eventMap_t::iterator it = pc->_pendingEvents.begin(); it != pc->_pendingEvents.end(); it++) {
            event *pEvent = it->first;
            cout<<pEvent->getProcess()->name();
            cout<<", ("<<pEvent->getAction()->getObject()->name();
            cout<<" :: "<<pEvent->getAction()->getName()<<"), ";
            cout<<astate[pEvent->getActionState()]<<" CS#="<<it->second;
            cout<<" RefCount="<<pEvent->getRefCount()<<endl;
        } // end for
        cout<<"}"<<endl;
        cout<<"in synch=";
        if (pc->_inSynch) cout<<"yes"<<endl;
        else cout<<"no"<<endl;

        cout<<"int funcType="<<pc->funcType<<" [ ";
        for (int j=1, k=1; j<=4; j++, k=k<<1) if (pc->funcType & k) cout<<ftype[j]<<" ";
        cout<<"]"<<endl;

        cout<<"bool * cond={";
        if (!pc->testList.empty()) {
            if (pc->cond) {
                _2DList_t *p = pc->testList.back();
                for (int j = 0; j < p->size(); j++) cout<<pc->cond[j]<<" ";
            } // end if
            cout<<"}"<<endl;
        } else
            cout<<"}"<<endl;

        cout<<"int selected="<<pc->selected<<endl;
        cout<<"csOrder={ ";
        for (int i = 0; i < pc->_csOrder.size(); i++) cout<<pc->_csOrder[i]<<" ";
        cout<<"}"<<endl;


        cout<<"testList={ ";
        for (int j = 0; j < pc->testList.size(); j++) {
            _2DList_t *pTestListStack = pc->testList[j];

            cout<<"#";
            for (int k = 0; k < pTestListStack->size(); k++) {
                simpleList_t *pTestList = (*pTestListStack)[k];

                cout<<"[";
                for (int m = 0; m < pTestList->size(); m++) {

                    node *pNode = (*pTestList)[m];
                    cout<<"("<<pNode->object->name()<<","<<pNode->intfcName.c_str()<<")";
                } // end for
                cout<<"]";
            } // end for
            cout<<"# ";
        } // end for
        cout<<"}"<<endl;


        cout<<"setList ={ ";
        for (int j = 0; j < pc->setList.size(); j++) {
            _2DList_t *pSetListStack = pc->setList[j];

            cout<<"#";
            for (int k = 0; k < pSetListStack->size(); k++) {
                simpleList_t *pSetList = (*pSetListStack)[k];

                cout<<"[";
                for (int m = 0; m < pSetList->size(); m++) {
                    node *pNode = (*pSetList)[m];

                    cout<<"("<<pNode->object->name()<<","<<pNode->intfcName.c_str()<<")";
                } // end for
                cout<<"]";
            } // end for
            cout<<"# ";
        } // end for
        cout<<"}"<<endl;

        cout<<"preventList={ ";
        for (int k = 0; k < pc->preventList.size(); k++) {
            simpleList_t *pPreventList = pc->preventList[k];

            cout<<"[";
            for (int m = 0; m < pPreventList->size(); m++) {
                node *pNode = (*pPreventList)[m];
                cout<<"("<<pNode->object->name()<<","<<pNode->intfcName.c_str()<<")";
            } // end for
            cout<<"]";
        } // end for
        cout<<"}"<<endl;

        cout<<"nodeList=[ ";
        for (int m = 0; m < pc->nodeList.size(); m++) {
            node *pNode = pc->nodeList[m];
            cout<<"("<<pNode->object->name()<<","<<pNode->intfcName.c_str()<<")";
        } // end for
        cout<<"]"<<endl;

        cout<<"labelList=[ ";
        for (int m = 0; m < pc->labelList.size(); m++)
            cout<<"("<<pc->labelList[m]->c_str()<<")";
        cout<<"]"<<endl;

        cout<<"SchedStateList=[ ";
        for (int m = 0; m < pc->SchedStateList.size(); m++)
            cout<<"("<<state[*(pc->SchedStateList[m])]<<")";
        cout<<"]"<<endl;

        cout<<"funcTypeList=[ ";
        for (int m = 0; m < pc->funcTypeList.size(); m++)
            cout<<"("<<ftype[*(pc->funcTypeList[m])]<<")";
        cout<<"]"<<endl;

        cout<<"potentialUsedMedia=[ ";
        for (MediaSet_t::iterator it = pc->_potentialUsedMedia->begin(); it!=pc->_potentialUsedMedia->end(); it++)
            cout<<(*it)->name()<<"("<<(*it)<<") ";
        cout<<"]"<<endl;

        cout<<"potentialUsedMediaList={ ";
        for (int m = 0; m<pc->_potentialUsedMediaList.size(); m++) {
            cout<<"#[";
            MediaSet_t * p = pc->_potentialUsedMediaList[m];
            for (MediaSet_t::iterator it = p->begin(); it!=p->end(); it++)
                cout<<(*it)->name()<<"("<<(*it)<<") ";
            cout<<"]# ";
        }
        cout<<"}"<<endl;

        cout<<endl;
    }

    cout<<"------ Media Information ------"<<endl;
    for (int i = 0; i < msize; i++) {
        medium *m = mediaVector[i];

        cout<<"Medium #"<<i+1<<"\t"<<m<<endl;
        cout<<"medium name="<<m->name()<<endl;
        cout<<"medium id="<<m->_id<<endl;

        cout<<"used interfaces:"<<endl;
        int idx = 1;
        for (interfaceUsers_t::iterator it=m->_intfcUsed.begin(); it!=m->_intfcUsed.end(); it++) {
            const char * intfcName = it->first;
            userPCs_t * pcs = it->second;
            cout<<idx<<". interface name:\t"<<intfcName<<endl;
            idx++;
            for (userPCs_t::iterator uit=pcs->begin(); uit!=pcs->end(); uit++) {
                ProgramCounter * pc = *uit;
                cout<<"\t"<<pc->p->name()<<"("<<pc<<")"<<endl;
            }
        }

        cout<<"prevented interfaces:"<<endl;
        idx = 1;
        for (interfaceUsers_t::iterator it=m->_intfcPrevented.begin(); it!=m->_intfcPrevented.end(); it++) {
            const char * intfcName = it->first;
            userPCs_t * pcs = it->second;
            cout<<idx<<". interface name:\t"<<intfcName<<endl;
            idx++;
            for (userPCs_t::iterator uit=pcs->begin(); uit!=pcs->end(); uit++) {
                ProgramCounter * pc = *uit;
                cout<<"\t"<<pc->p->name()<<"("<<pc<<")"<<endl;
            }
        }

        cout<<"potential users:"<<endl;
        for (potentialUserPCs_t::iterator it=m->_potentialUserPCs.begin(); it!=m->_potentialUserPCs.end(); it++) {
            ProgramCounter * pc = *it;
            cout<<"\t"<<pc->p->name()<<"("<<pc<<")"<<endl;
        }

        cout<<endl;
    }

    cout<<"------ Synch Events Information ------"<<endl;
    cout<<"Number of synch event groups = "<<numSynchEventGroup<<endl;
    for (int i=0; i<numSynchEventGroup; i++) {
        cout<<"Group "<<i<<":"<<endl;
        cout<<"\tNumber of synch events "<<numSynchEventEachGroup[i]<<endl;
        cout<<"\tEnabled synch events "<<enabledSynchEventEachGroup[i]<<endl;
        cout<<"\tEnabled synch events PCs"<<endl;
        for (int j=0; j<enabledSynchEventEachGroup[i]; j++) {
            ProgramCounter *pc = enabledSynchEventPC[i][j];
            cout<<"\t\t"<<pc->p->name()<<"("<<pc<<")"<<endl;
        }
    }
    cout<<endl;

#ifdef DEBUG_INTERACTIVE
    char c;
    cin >> c;
#endif
}

void scoreboard::invoke() {

    bool flag = false;

    for (int i = 0; i < size; i++) {

        if (testSchedFlag) {
            if (pcVector[i]->SchedState == SchedStateVal::RUN ||
                    pcVector[i]->mode == RunType::TRY ||
                    pcVector[i]->_inSynch ) {
                flag = true;
                if (timestat) NumRunProcesses++;
                notify(DeltaCycle, pcVector[i]->invoke);
                if (debug_flag) {
                    cout<<"Invoke process "<<pcVector[i]->p->name()<<" ("<<pcVector[i]<<")"<<endl;
                }
            }
        } else {

            if ((pcVector[i]->funcType & FunctionType::AWAIT) ||
                    (pcVector[i]->funcType & FunctionType::ANNOTATION) ||
                    (pcVector[i]->SchedState == SchedStateVal::RUN) ||
                    (pcVector[i]->_inSynch) )

                {
                    flag = true;
                    if (timestat)
                        if (pcVector[i]->SchedState == SchedStateVal::RUN)
                            NumRunProcesses++;
                    notify(DeltaCycle, pcVector[i]->invoke);
                    if (debug_flag) {
                        cout<<"Invoke process "<<pcVector[i]->p->name()<<" ("<<pcVector[i]<<")"<<endl;
                    }
                }
        } // end if

    } // end for

    if (!flag) {
        cout<<"Finished or Deadlocked!"<<endl;
        //showSB();
        sc_sim_stop();
    } // end if

    wait(_mng.pc->invoke);

}

//Fix: should support more than two synched events
void scoreboard::registerSynchEvents(event *e1, event *e2) {
    _eventSynchMap.insert(eventSynchMap_t::value_type(new EventMap(e1), new EventMap(e2)));
} // end registerSynchEvents

EventMap *scoreboard::getEventSynchMap(event *e, bool pair) {
    EventMap *pResult = NULL;

    // search this event in the synch map
    for (eventSynchMap_t::iterator it = _eventSynchMap.begin(); !pResult && it != _eventSynchMap.end(); it++) {
        EventMap *pEM_1 = it->first;
        EventMap *pEM_2 = it->second;
        if (pEM_1->_e == e) pResult = pair ? pEM_2 : pEM_1;
        if (pEM_2->_e == e) pResult = pair ? pEM_1 : pEM_2;
    } // end for

    return pResult;
} // end getEventSynchMap

//bool scoreboard::atLeastOneRun() {
//  	for (int i = 0; i < size; i++)
//        if (pcVector[i]->SchedState == SchedStateVal::RUN) return true;
//     return false;
//} // end atLeastOneRun

// Initialize SAT solver ZChaff
void scoreboard::initSATSolver() {
    SatMgr = SAT_InitManager();
    if (heuristic == 4) SAT_AddHookFun(SatMgr, favor_1_sat_hook, 1);

} // end initSATSolver

// Free SAT solver ZChaff
void scoreboard::freeSATSolver() {
    SAT_ReleaseManager(SatMgr);
} // end freeSATSolver

// Create variables in and get their indices from ZChaff for all events in eventVarIndex
void scoreboard::createVars(eventVarIndexMap_t& evi) {
    if (evi.empty()) return;
    int index;
    for (eventVarIndexMap_t::iterator it=evi.begin(); it!=evi.end(); it++) {
        index = SAT_AddVariable(SatMgr);
        it->second = index;
    } // end for
} // end createVars

// Add synch constraints as CNF's to ZChaff
void scoreboard::addSynchClauses(eventVarIndexMap_t& evi, eventSynchMap_t& esm) {
    int clause[2];
    for (eventSynchMap_t::iterator it=esm.begin(); it!=esm.end(); it++) {
        clause[0] = POSLIT(evi[it->first->_e]);
        clause[1] = NEGLIT(evi[it->second->_e]);
        SAT_AddClause(SatMgr, clause, 2);
        clause[0] ++; //change to negative literal
        clause[1] --;   //change to positive literal
        SAT_AddClause(SatMgr, clause, 2);
    } // end for
} // end addSynchClauses

// Add conflict constraints stored in conflictRSet as CNF's to ZChaff
void scoreboard::addConflictClauses(eventVarIndexMap_t& evi, conflictRelationSet_t& crs) {
    int clause[2];
    for (conflictRelationSet_t::iterator it=crs.begin(); it!=crs.end(); it++) {
        clause[0] = NEGLIT(evi[it->first]);
        clause[1] = NEGLIT(evi[it->second]);
        SAT_AddClause(SatMgr, clause, 2);
    } // end for
} // end addConflictClauses

// Add number (lower bound) of 1-value variable constraints as CNF's to ZChaff
void scoreboard::addNumberOfOnesClauses(eventVarIndexMap_t::iterator it,
        int size, int need, intVect_t& result) {
    if (need == 0) {
        int ss = result.size();
        if (ss==0) return;
        int *clause = new int[ss];
        for (int i=0; i<ss; i++) clause[i] = result[i];
        SAT_AddClause(SatMgr, clause, ss);
        //    SAT_AddClause(SatMgr, result.begin(), ss);
        delete clause;
        return;
    } // end if

    result.push_back(POSLIT(it->second));
    addNumberOfOnesClauses(++it, size-1, need-1, result);
    --it;
    result.pop_back();
    if (size>need) addNumberOfOnesClauses(++it, size-1, need, result);
    --it;
} // end addNumberOfOnesClauses

// Add LTL constraints coming from transition relations in Buchi Automaton
void scoreboard::addLTLConstraintClauses(eventVarIndexMap_t& evi) {
    topNetlist->addLTLConstraintsToSAT(SatMgr, evi);
} // end addLTLConstraintClauses


// Solve the SAT problem by calling ZChaff
enum SAT_StatusT  scoreboard::solveSAT() {
    return (enum SAT_StatusT) SAT_Solve(SatMgr);
} // end solveSAT

// Get the result from SAT solver
void scoreboard::getSATResult(eventVarIndexMap_t& evi) {
    for (eventVarIndexMap_t::iterator it=evi.begin(); it!=evi.end(); it++) {
        evi[it->first] = SAT_GetVarAsgnment(SatMgr, it->second);
    } //end for
} // end getSATResult

// Apply SAT result to the scheduling
bool scoreboard::applySATResult(eventVarIndexMap_t& evi) {
    bool flag = false;
    for (eventVarIndexMap_t::iterator it=evi.begin(); it!=evi.end(); it++) {
        if (it->second==0) // variable is 0 => corresponding event is disabled
            it->first->getProcess()->pc->setMustNotDo(it->first);
        else
            flag = true;
    } //end for
    return flag;
} // end applySATResult

// Count the number of 1-value variables
int scoreboard::countOnes(eventVarIndexMap_t& evi) {
    int numZeros = 0;
    for (eventVarIndexMap_t::iterator it=evi.begin(); it!=evi.end(); it++) {
        if (it->second==0) numZeros++; // variable is 0
    } //end for
    return (evi.size() - numZeros);
} // end countOnes

// Randomly choose one event and set it (and its synched event, if any) to run if possible
bool scoreboard::pickOneEventRandomly() {
    int * pickOrder;
    pickOrder = intArrayGen(size, true);
    bool picked = false;
    ProgramCounter *pc, *mpc;
    pc = mpc = NULL;

    for (int i=0; i<size && !picked; i++) {
        pc = pcVector[pickOrder[i]];
        if (testSchedFlag) {
            if (pc->SchedState != SchedStateVal::TESTSCHEDULING) continue;
        } else {
            if (pc->SchedState != SchedStateVal::UNKNOW) continue;
        } // end if

        int psize;
        if ((psize=pc->_pendingEvents.size()) == 0) continue;

        for (; psize>0; psize--) {
            int choice = (int)(random() % psize);
            eventMap_t::iterator it = pc->_pendingEvents.begin();
            for (int ii = 0 ; it != pc->_pendingEvents.end() && ii < choice; it++, ii++);
            event * e = it->first;
            pc->selected = it->second;

            EventMap * emap = getEventSynchMap(e, true);
            if (emap==NULL) {
                picked = true;
                break;
            } else {
                event *em = emap->_e;
                mpc = em->getProcess()->pc;
                if ((testSchedFlag && mpc->SchedState != SchedStateVal::TESTSCHEDULING) ||
                        (!testSchedFlag && mpc->SchedState != SchedStateVal::UNKNOW)) {
                    pc->setMustNotDo(e);
                    continue;
                } else {
                    pc->setMustDo(e);
                    mpc->setMustDo(em);
                    picked = true;
                    break;
                } // end if
            } // end if
        } // end for
    } // end for

    // make all other processes not to run
    for (int i=0; i<size; i++) {
        if ((pcVector[i]==pc) || (pcVector[i]==mpc)) continue;
        pcVector[i]->_pendingEvents.clear();
    } // end for


    return picked;
} // end pickOneEventRandomly

bool scoreboard::schedule() {
    long tmp;
    struct rusage ru;
    if (timestat) {
        NumSMSchedulingRounds ++;
        getrusage(RUSAGE_SELF, &ru);
  	tmp = ru.ru_utime.tv_sec*1000000+ru.ru_utime.tv_usec+
            ru.ru_stime.tv_sec*1000000+ru.ru_stime.tv_usec;
    } // end if

    bool flag = (this->*SchedAlg)();

    if (timestat) {
        getrusage(RUSAGE_SELF, &ru);
  	tmp = ru.ru_utime.tv_sec*1000000+ru.ru_utime.tv_usec+
            ru.ru_stime.tv_sec*1000000+ru.ru_stime.tv_usec - tmp;
        SMSchedulingTime += tmp;
    } // end if

    return flag;
}

// Scheduling algorithm 0: the original pass1-2-3... algorithm
// This one is the default algorithm used.
bool scoreboard::heuristic0() {
    bool flag;

    if (ANNOTATION_ENABLED) {
        topNetlist->resolve((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after resolving quantity managers"<<endl;
            showSB();
        }
    }

    if (ltl) {
        topNetlist->preGuideLTL();
        if (debug_flag && debug_level>3) {
            cout<<endl<<"after preGuideLTL"<<endl;
            showSB();
        }
    }

    flag = pass1();
    if (debug_flag && debug_level>1) {
        cout<<endl<<"after pass1"<<endl;
        showSB();
    }

    if (flag) {
        flag = pass2();
        if (debug_flag && debug_level>1) {
            cout<<endl<<"after pass2"<<endl;
            showSB();
        }
    }

    /*
      if (flag) {
      flag = pass3();
      if (debug_flag && debug_level>4) {
      cout<<endl<<"after pass3"<<endl;
      showSB();
      }
      }
    */

    if (flag) {
        flag = resolveSynch();
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after resolveSynch"<<endl;
            showSB();
        }
    }

    if (ANNOTATION_ENABLED) {
        topNetlist->postcond((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after running postcond of quantity managers"<<endl;
            showSB();
        }
    }

    event * passLTL = ((event*) 1); // NULL:  passed; -1: violated; other: need to iterate
    flag = true;
    if (ltl) {
        while ((passLTL!=NULL) && (passLTL!=((event*)-1)) && flag ) {
            flag = set2run();
            passLTL = topNetlist->postCheckLTL(true);
        } // end while
    } else {
        flag = set2run();
    } // end if

    if (debug_flag && debug_level>3) {
        cout<<endl<<"after set2run";
        if (ltl) cout<<" and postCheckLTL";
        cout<<endl;
        showSB();
    }

    return flag;
} // end heuristic0

// Scheduling algorithm 1: randomly choose one process to run
bool scoreboard::heuristic1() {
    bool flag;

    if (ANNOTATION_ENABLED) {
        topNetlist->resolve((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after resolving quantity managers"<<endl;
            showSB();
        }
    }

    if (ltl) {
        topNetlist->preGuideLTL();
        if (debug_flag && debug_level>3) {
            cout<<endl<<"after preGuideLTL"<<endl;
            showSB();
        }
    }

    flag = pass1();
    if (debug_flag && debug_level>1) {
        cout<<endl<<"after pass1"<<endl;
        showSB();
    }

    event * passLTL = ((event*) 1); // NULL:  passed; -1: violated; other: need to iterate
    if (flag && ltl) {
        while ((passLTL!=NULL) && (passLTL!=((event*)-1))) {
            passLTL = topNetlist->postCheckLTL(false);
        } // end while
    } // end if

    if (flag) {
        flag = pickOneEventRandomly();
        if (debug_flag && debug_level>1) {
            cout<<endl<<"after randomly pick one event"<<endl;
            showSB();
        }
    }

    if (ANNOTATION_ENABLED) {
        topNetlist->postcond((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after running postcond of quantity managers"<<endl;
            showSB();
        }
    }

    passLTL = ((event*) 1);
    if (ltl && flag) {
        while ((passLTL!=NULL) && (passLTL!=((event*)-1)) && flag ) {
            flag = set2run();
            passLTL = topNetlist->postCheckLTL(true);
        } // end while
    } else {
        flag = set2run();
    } // end if

    if (debug_flag && debug_level>3) {
        cout<<endl<<"after set2run";
        if (ltl) cout<<" and postCheckLTL";
        cout<<endl;
        showSB();
    }

    return flag;
} // end heuristic1

// Scheduling algorithm 2: call zchaff to find a set of consistent processes to run
bool scoreboard::heuristic2() {
    bool flag;

    if (ANNOTATION_ENABLED) {
        topNetlist->resolve((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after resolving quantity managers"<<endl;
            showSB();
        }
    }

    if (ltl) {
        topNetlist->preGuideLTL();
        if (debug_flag && debug_level>3) {
            cout<<endl<<"after preGuideLTL"<<endl;
            showSB();
        }
    }

    flag = pass1();
    if (debug_flag && debug_level>1) {
        cout<<endl<<"after pass1"<<endl;
        showSB();
    }

    if (flag) {
        // collect all pending events, and the confliction among them
        collectEventConflicts();

        // Initialize SAT solver ZChaff
        initSATSolver();

        // Create variables in and get their indices from ZChaff for all events in eventVarIndex
        createVars(eventVarIndex);

        // Add synch constraints as CNF's to ZChaff
        addSynchClauses(eventVarIndex, _eventSynchMap);

        // Add conflict constraints stored in conflictRSet as CNF's to ZChaff
        addConflictClauses(eventVarIndex, conflictRSet);

        // At least one event should run. Add this constraints as CNF's to ZChaff
        intVect_t dummy;
        addNumberOfOnesClauses(eventVarIndex.begin(), (int)eventVarIndex.size(), (int)eventVarIndex.size(), dummy);

        // Add LTL constraints coming from transition relations in Buchi Automaton
        if (ltl) addLTLConstraintClauses(eventVarIndex);

        // Solve the SAT problem by calling ZChaff
        switch(solveSAT()) {
        case UNSATISFIABLE:
            cerr<<"Unable to finish simulation, because your design has internal conflicts."<<endl;
            sc_stop();
            break;
        case SATISFIABLE:
            break;
        default:
            cerr<<"Unable to finish simulation, because the underline SAT engine cannot handle the problem."<<endl;
            sc_stop();
        } // end switch

        // Get the result from SAT solver
        getSATResult(eventVarIndex);

        // Apply SAT result to the scheduling
        flag = applySATResult(eventVarIndex);

        // Free SAT solver ZChaff
        freeSATSolver();

        if (debug_flag && debug_level>1) {
            cout<<endl<<"after applySATResult"<<endl;
            showSB();
        }

    } // end if

    if (ANNOTATION_ENABLED) {
        topNetlist->postcond((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after running postcond of quantity managers"<<endl;
            showSB();
        }
    }

    event * passLTL = ((event*) 1); // NULL:  passed; -1: violated; other: need to iterate
    flag = true;
    if (ltl) {
        while ((passLTL!=NULL) && (passLTL!=((event*)-1)) && flag ) {
            flag = set2run();
            passLTL = topNetlist->postCheckLTL(true);
        } // end while
    } else {
        flag = set2run();
    } // end if

    if (debug_flag && debug_level>3) {
        cout<<endl<<"after set2run";
        if (ltl) cout<<" and postCheckLTL";
        cout<<endl;
        showSB();
    }

    return flag;
} // end heuristic2

// Scheduling algorithm 3: iteratively call zchaff to find a MAX set of consistent processes to run
bool scoreboard::heuristic3() {
    bool flag;

    if (ANNOTATION_ENABLED) {
        topNetlist->resolve((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after resolving quantity managers"<<endl;
            showSB();
        }
    }

    if (ltl) {
        topNetlist->preGuideLTL();
        if (debug_flag && debug_level>3) {
            cout<<endl<<"after preGuideLTL"<<endl;
            showSB();
        }
    }

    flag = pass1();
    if (debug_flag && debug_level>1) {
        cout<<endl<<"after pass1"<<endl;
        showSB();
    }

    if (flag) {
        // collect all pending events, and the confliction among them
        collectEventConflicts();

        bool foundMaxOnes = false;
        int numOfOnes = 0;
        eventVarIndexMap_t result;

        while (!foundMaxOnes) {

            // Initialize SAT solver ZChaff
            initSATSolver();

            // Create variables in and get their indices from ZChaff for all events in eventVarIndex
            createVars(eventVarIndex);

            // Add synch constraints as CNF's to ZChaff
            addSynchClauses(eventVarIndex, _eventSynchMap);

            // Add conflict constraints stored in conflictRSet as CNF's to ZChaff
            addConflictClauses(eventVarIndex, conflictRSet);

            // Add number (lower bound) of 1-value variable constraints as CNF's to ZChaff
            intVect_t dummy;
            addNumberOfOnesClauses(eventVarIndex.begin(), (int)eventVarIndex.size(), eventVarIndex.size()-numOfOnes, dummy);

            // Add LTL constraints coming from transition relations in Buchi Automaton
            if (ltl) addLTLConstraintClauses(eventVarIndex);

            // Solve the SAT problem by calling ZChaff
            NumIterH3 ++;
            switch(solveSAT()) {
            case UNSATISFIABLE:
                if (numOfOnes == 0) {
                    cerr<<"Unable to finish simulation, because your design has internal conflicts."<<endl;
                    sc_stop();
                } else
                    foundMaxOnes = true;
                break;
            case SATISFIABLE:
                // Get SAT result from SAT solver
                getSATResult(eventVarIndex);

                // Count the number of 1-value variables
                int tmp;
                tmp = countOnes(eventVarIndex);

                if (tmp>numOfOnes) {
                    numOfOnes = tmp;
                    result = eventVarIndex; // Save current SAT result
                    if (numOfOnes+1>result.size()) foundMaxOnes = true;
                } else
                    foundMaxOnes = true;
                break;
            default:
                if (numOfOnes == -1) {
                    cerr<<"Unable to finish simulation, because the underline SAT engine cannot handle the problem."<<endl;
                    sc_stop();
                } else
                    foundMaxOnes = true;
            } // end switch

            // Free SAT solver ZChaff
            freeSATSolver();

        } // end while

        // Apply SAT result to the scheduling
        flag = applySATResult(result);
        if (debug_flag && debug_level>1) cout<<endl<<"Total number of scheduled events is "<<countOnes(result)<<endl;

        if (debug_flag && debug_level>1) {
            cout<<endl<<"after applySATResult"<<endl;
            showSB();
        }

    } // end if

    if (ANNOTATION_ENABLED) {
        topNetlist->postcond((process*)&_mng);
        if (debug_flag && debug_level>2) {
            cout<<endl<<"after running postcond of quantity managers"<<endl;
            showSB();
        }
    }

    event * passLTL = ((event*) 1); // NULL:  passed; -1: violated; other: need to iterate
    flag = true;
    if (ltl) {
        while ((passLTL!=NULL) && (passLTL!=((event*)-1)) && flag ) {
            flag = set2run();
            passLTL = topNetlist->postCheckLTL(true);
        } // end while
    } else {
        flag = set2run();
    } // end if

    if (debug_flag && debug_level>3) {
        cout<<endl<<"after set2run";
        if (ltl) cout<<" and postCheckLTL";
        cout<<endl;
        showSB();
    }

    return flag;
} // end heuristic3

// Used by the metamodel debugger.
// Uses a std::set to eliminate duplicate strings.
//
strList_t scoreboard::getAllNextLines() {
    strList_t* nextLineList = new strList_t;
    std::set<String> lineSet;
    std::set<String>::iterator setIter;

    for (int i=0; i<size; i++) {
	ProgramCounter* pc = pcVector[i];
	for (int j=0; j<pc->nextMMLines.size(); j++) {
	    lineSet.insert(pc->nextMMLines[j]);
	}
    }

    for (setIter = lineSet.begin(); setIter != lineSet.end(); setIter++) {
	String* s = new String(*setIter);
	nextLineList->push_back(s);
    }
    return *nextLineList;
}

// Used by the metamodel debugger.
//
strList_t scoreboard::getAllCurrentLines() {
    // strList_t* currentLineList = new strList_t;
    /*
      currentLineList = new strList_t;

      for (int i = 0; i < size; i++) {
      ProgramCounter* pc = pcVector[i];
      String *pCurrentLineString = new String(pc->p->name());
      if (pCurrentLineString != (String*)0) {
      String currentLineString = *pCurrentLineString;
      if (pc->currentLineNumber != (String *)0) {
      currentLineString += ":";
      currentLineString += *pc->currentLineNumber;
      currentLineList->push_back(new String(currentLineString));
      }
      }
      }
    */
    updateCurrentLines();

    /*
     * "Manager" process line number is not of interest to mgdb users.
     * String *pCurrentLineString = new String(_mng.name());
     * if (pCurrentLineString != (String*)0) {
     *     String currentLineString = *pCurrentLineString;
     *     if (_mng.pc->currentLineNumber != (String *)0) {
     *         currentLineString += ":";
     *         currentLineString += *_mng.pc->currentLineNumber;
     *         currentLineList->push_back(new String(currentLineString));
     *     }
     * }
     */

    return *currentLineList;
}

// Used by the metamodel debugger.
//
strList_t scoreboard::getAllEventTags() {
    strList_t* tagList = new strList_t;

    for (int i = 0; i < size; i++) {
        ProgramCounter* pc = pcVector[i];
        String *pTagString = new String(pc->p->name());
        if (pTagString != (String*)0) {
            String tagString = *pTagString;
            if (pc->eventTag != (String *)0) {
                tagString += ":";
                tagString += *pc->eventTag;
                tagList->push_back(new String(tagString));
            }
        }
    }
    return *tagList;
}


// Used by the metamodel debugger.
//
strList_t scoreboard::getAllSchedStates() {
    strList_t* stateList = new strList_t;

    for (int i = 0; i < size; i++) {
        ProgramCounter* pc = pcVector[i];
        String *pStateString = new String(pc->p->name());
        if (pStateString != (String*)0) {
            String stateString = *pStateString;
            stateString += ":";
            switch (pc->SchedState) {
            case SchedStateVal::RUN:
                stateString += "RUN";
                break;
            case SchedStateVal::DONTRUN:
                stateString += "DONTRUN";
                break;
            case SchedStateVal::RUNNING:
                stateString += "RUNNING";
                break;
            case SchedStateVal::UNKNOW:
                stateString += "UNKNOWN";
                break;
            case SchedStateVal::EVALUATE:
                stateString += "EVALUATE";
                break;
            case SchedStateVal::FINISHED:
                stateString += "FINISHED";
                break;
            case SchedStateVal::TESTSCHEDULING:
                stateString += "TESTSCHEDULING";
                break;
            default:
                stateString += "???";
                break;
            }
            stateList->push_back(new String(stateString));
        }
    }
    return *stateList;
}

//
void scoreboard::updateCurrentLines() {
    currentLineList = new strList_t;

    for (int i = 0; i < size; i++) {
        ProgramCounter* pc = pcVector[i];
        String *pCurrentLineString = new String(pc->p->name());
        if (pCurrentLineString != (String*)0) {
            String currentLineString = *pCurrentLineString;
            if (pc->currentLineNumber != (String *)0) {
                currentLineString += ":";
                currentLineString += *pc->currentLineNumber;
                currentLineList->push_back(new String(currentLineString));
            }
        }
    }
}
