/* A constraint in the elaborated network.

Metropolis: Design Environment for Heterogeneus Systems.

Copyright (c) 1998-2004 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in all
copies of this software and that appropriate acknowledgments are made
to the research of the Metropolis group.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

*/


package metropolis.metamodel.runtime;

import java.io.Serializable;

//////////////////////////////////////////////////////////////////////////
//// Event
/**
   An event in the elaborated network.

   @author   Xi Chen, Contributor: Christopher Hylands Brooks
   @version  $Id: Event.java,v 1.21 2004/05/20 04:18:10 cxh Exp $
*/
public class Event implements Serializable {

    ///////////////////////////////////////////////////////////////////
    ////                           constructors                    ////

    /** Build a new event in the network.

    *  <p>Events are one of beg(process, object.name), end(process,
    *  object.name), none(process), or other(process)
    *
    *  @param kind The kind of event, one of
    *  {@link #BEG}, {@link #END}, {@link #NONE}, or {@link #OTHER}.
    *  @param process The process where the event belongs to. It is
    *  "null" if the process is set to "all".
    *  @param nodeObject The object node of the action. It is "null"
    *  if kind is NONE or OTHER or if the action is set to "all"
    *  @param name The label or function name in the action. It is "null"
    *  if kind is NONE or OTHER, or if the action is set to "all"
    */

    public Event(int kind, Object process, Object nodeObject, String name) {
        // FIXME: Should the type of process be a INode?
        _kind        = kind;
        if (_kind < BEG || _kind > OTHER) {
            throw new RuntimeException("kind '" + _kind + "'was out of range, "
                    + "it must be >= " + BEG + " and <= " + OTHER);
        }
        _process     = null;
        _nodeObject  = null;
        _name        = name;

        // We used to catch any errors and then exit, but this is very harsh
        // and makes it hard to test the code.
        _process = Network.net.getNode(process);
        _nodeObject = Network.net.getNode(nodeObject);

    }
    ///////////////////////////////////////////////////////////////////
    ////

    // FIXME: In Java 1.5, use an enum.

    /** Kind for beg(process, object.name). */
    public static final int BEG   = 0;

    /** Kind for end(process, object.name). */
    public static final int END   = 1;

    /** Kind for none(process). */
    public static final int NONE  = 2;

    /** Kind for other(process). */
    public static final int OTHER = 3;

    ///////////////////////////////////////////////////////////////////
    ////                          public methods                   ////

    /** See if this event is the same as another event.
     *  @param event The event that is compared to this event.
     *  @return If is the same return true, otherwise return false.
     */
    public boolean equals(Event event) {
        // FIXME: Need a hashcode()
        // http://java.sun.com/docs/books/tutorial/java/javaOO/objectclass.html
        if (event == null) {
            return false;
        }

        // FIXME: what if event.getName() is null?
        if (event.getKind() == _kind
                && event.getProcess() == _process
                && event.getNodeObject() == _nodeObject
                && event.getName().equals(_name)) {
            return true;
        } else {
            return false;
        }
    }

    /** Get the process of this event.
     *  @return The process node of this connection.
     */
    public INode getProcess() {
        return _process;
    }

    /** Get the kind of this event.
     *  @return The kind of this event.
     */
    public int getKind() {
        return _kind;
    }

    /** Get the label or function name of the action of the event.
     *  @return The label or function name of the action of the event.
     */
    public String getName() {
        return _name;
    }

    /** Get the object node of the action of the event.
     *  @return The object node of the action of the event.
     */
    public INode getNodeObject() {
        return _nodeObject;
    }

    /** Get a String providing information about this event.
     *  @return A String providing information about this event.
     */
    public String show() {
        String returnString = null;
        String processName = null;
        String actionName = null;

        if (_process == null) {
            processName = "all";
        } else {
            processName = _process.getName();
        }

        if (_nodeObject == null) {
            actionName = "all";
        } else {
            actionName = _nodeObject.getName() + "." + _name;
        }

        switch(_kind) {
        case BEG:
            returnString = "beg(" + processName + ", " + actionName + ")";
            break;
        case END:
            returnString = "end(" + processName + ", " + actionName + ")";
            break;
        case NONE:
            returnString = "none(" + processName + ")";
            break;
        case OTHER:
            returnString = "other(" + processName + ")";
            break;
        default:
            throw new RuntimeException("Event.show(), unexpected event '"
                    + _kind + "'");
            //break;
        }
        return  returnString;
    }

    /** Get a String representing the event,
     *  e.g. "BEG_processName_nodeObject_label".
     *  @return A String representing the event.
     */
    public String toString() {
        String returnString = new String();
        String processName = new String();
        String actionName = new String();

        if (_process == null) {
            return returnString;
        } else {
            processName = _process.getName();
        }

        if (_nodeObject == null) {
            return returnString;
        } else {
            actionName = _nodeObject.getName() + "_" + _name;
        }

        switch(_kind) {
        case 0:
            returnString = "BEG_" + processName + "_" + actionName;
            break;
        case 1:
            returnString = "END_" + processName + "_" + actionName;
            break;
        }
        return  returnString;
    }

    ///////////////////////////////////////////////////////////////////
    ////                        protected variables                ////

    /** Kind of event. One of
     *  {@link #BEG}, {@link #END}, {@link #NONE}, or {@link #OTHER}.
     */
    protected int _kind;

    /** Process of event. */
    protected INode _process;

    /** Object of the action in the event. */
    protected INode _nodeObject;

    /** Label or function name of the action in the event. */
    protected String _name;

}
