/* Elaborated network of a meta-model specification.

Metropolis: Design Environment for Heterogeneus Systems.

Copyright (c) 1998-2004 The Regents of the University of California.
All rights reserved.

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the
above copyright notice and the following two paragraphs appear in all
copies of this software and that appropriate acknowledgments are made
to the research of the Metropolis group.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

*/


package metropolis.metamodel.runtime;

import metropolis.metamodel.StringManip;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

//////////////////////////////////////////////////////////////////////////
//// Network
/**
 * This class describes the elaborated network in the meta-model
 * specification, and offers methods to implement meta-model constructs
 * easily. For example, redirectconnect is implemented by method
 * redirectConnect(), etc.
 *
 * @author   Robert Clariso
 * @version  $Id: Network.java,v 1.59 2004/05/20 04:41:19 cxh Exp $
 */
public class Network implements Serializable {


    ///////////////////////////////////////////////////////////////////
    ////                         public variables                  ////

    /** Current instance of the network that we are working with. */
    public static Network net = new Network();


    ///////////////////////////////////////////////////////////////////
    ////                          public methods                   ////

    /** Get the instance name of a network node.
     *  @param instance A network node.
     *  @return The name of the instance.
     */
    public String getInstName(Object instance) {
        INode inst = getNode(instance);
        // FIXME: what if getNode() returns null?
        return getInstName(inst);
    }

    /** Gets this object's inst name.
     *  @param instance  A network node.
     *  @return The name of the instance.
     */
    public String getInstName(INode instance) {
        if (instance != null) {
            return instance.getName();
        } else {
            return "null";
        }
    }

    /** Get the component name of a network node in a netlist.
     *  @param instance A network node.
     *  @param netlist  A netlist node.
     *  @return The compName.
     */
    public String getCompName(Object instance, Object netlist) {
        INode inst = null;
        INode nl = null;
        inst = getNode(instance);
        nl = getNode(netlist);
        return getCompName(inst, nl);
    }

    /** Get the component name of a network node in a netlist.
     *  @param instance A network node.
     *  @param netlist  A netlist node.
     *  @return The compName.
     */
    public String getCompName(INode instance, INetlist netlist) {
        if (instance != null) {
            return instance.getCompName(netlist);
        } else {
            return "null";
        }
    }

    /** Get the component name of a network node in a netlist.
     *  @param nodeName Name of the network node.
     *  @param netName  Name of the netlist.
     *  @return The compName.
     */
    public String getCompName(String nodeName, String netName) {
        INode instance = getNode(nodeName);
        INetlist netlist = getNetlist(netName);

        if (instance != null && netlist != null) {
            return instance.getCompName(netlist);
        } else {
            return "null";
        }
    }

    /** Get the type of  a node.
     *  @param instance The node to get the instance of
     *  @return The type
     */
    public MMType getType(INode instance) {

        return instance.getType();
    }

    /** Test if a connection through port from src to dest is refined or not.
     *  @param srcName Name of source node.
     *  @param portName Port name.
     *  @param destName Name ofdestination node.
     *  @return True if a connection is refined.
     */
    public boolean isConnectionRefined(String srcName, String portName,
            String destName) {
        INode src = getNode(srcName);
        INode dest = getNode(destName);
        IPort port = src.getPort(portName);

        return isConnectionRefined(src, port, dest);
    }

    /** Test if a connection through port from src to dest is refined or not.
     *  @param src source node.
     *  @param port Port.
     *  @param dest destination node.
     *  @return True if a connection is refined.
     *  @exception RuntimeException if there is no valid connection from
     *  src to dest through port
     */
    public boolean isConnectionRefined(INode src, IPort port, INode dest) {
        if (src == null || port == null || dest == null) {
            throw new RuntimeException("isConnectionRefined: "+
                    "none of src, port or dest could be null");
        }

        // Check that the source port belongs to the source object
        if (port.getContainer() != src) {
            throw new RuntimeException("isConnectionRefined: port '"
                    + port.show()
                    + "' does not belong to node '" + src.getName()
                    + "', but to node '" + port.getContainer().getName()
                    + "'");
        }

        // Check if there is connections from the source node to dest node that
        // uses the port.
        Iterator iterator = src.getPortConnections(port).iterator();
        Connection conn = null;
        while (iterator.hasNext()) {
            Connection tmpConn = (Connection)iterator.next();
            if (tmpConn.getTarget() == dest) {
                conn = tmpConn;
                break;
            }
        }

        if (conn == null) {
            throw new RuntimeException("isConnectionRefined: connection "+
                    "does not exist, there is no connection from node '"
                    + src.getName() + "' using port '" + port.show()
                    + "' to node " + dest.getName());
        }

        return conn.isRefined();
    }

    /** Get the index-th port of type ifName defined in node.
     *  @param srcObject source node.
     *  @param ifName name of interface.
     *  @param index the index of the port.
     *  @return The port.
     */
    public Object getNthPort(Object srcObject, String ifName, int index) {
        // FIXME: Should the return type be something other than Object?
        INode src = null;
        try {
            src = getNode(srcObject);
        } catch (Exception e) {
            System.err.println("getnthport: " + e.toString());
            System.exit(1);
        }
        return getNthPort(src, ifName, index);
    }

    /** Get the index-th port of type ifName defined in node.
     *  @param node network node.
     *  @param ifName name of interface.
     *  @param index the index of the port.
     *  @return The port.
     */
    public Object getNthPort(INode node, String ifName, int index) {
        // FIXME: Should the return type be something other than Object?
        MMType nodeType = node.getType();
        int count = index;
        Object obj = node.getUserObject();
        Class cls = obj.getClass();
        Iterator iterator = nodeType.portsIterator();
        while (iterator.hasNext()) {
            MMPort portType = (MMPort)iterator.next();
            if (portType.getInterface().getName().endsWith(ifName)) {
                Field f = null;
                while (cls != null && f==null) {
                    try {
                        f = cls.getDeclaredField(portType.getName());
                    } catch (NoSuchFieldException e) {
                        cls = cls.getSuperclass();
                    }
                }

                if (f == null) {
                    System.err.println("getnthport: no field " + portType.getName() + " of type ifName declared.");
                    System.exit(1);
                }

                try {
                    f.setAccessible(true);
                    if (f.getType().isArray()) {
                        int length = Array.getLength(f.get(obj));
                        if (count<length)
                            return Array.get(f.get(obj), count);
                        else
                            count -= length;
                    } else {
                        if (count==0)
                            return f.get(obj);
                        count --;
                    }
                } catch (IllegalAccessException e) {
                    System.err.println("getnthport: " + e.toString());
                    System.exit(1);
                }
            } // end if
        } // end while iterator

        return null;
    }

    /** Get the index-th port of type ifName defined in node.
     *  @param nodeName name of a network node.
     *  @param ifName name of interface.
     *  @param index the index of the port.
     *  @return The port.
     */
    public Object getNthPort(String nodeName, String ifName, int index) {
        INode node = getNode(nodeName);
        return getNthPort(node, ifName, index);
    }

    /** Get the number of elementary ports of type ifName defined in node.
     *  @param node network node.
     *  @param ifName name of interface.
     *  @return The port number.
     */
    public int getPortNum(INode node, String ifName) {
        if (node == null)
            return 0;
        MMType nodeType = node.getType();
        Iterator iterator = nodeType.portsIterator();
        int count = 0;
        while (iterator.hasNext()) {
            MMPort portType = (MMPort)iterator.next();
            if (portType.getInterface().getName().equals(ifName)) {
                count += node.getPort(portType.getName()).numPorts();
            }
        }

        return count;

    }

    /**  Get the number of elementary ports of type ifName defined in node.
     *  @param nodeName name of network node.
     *  @param ifName name of interface.
     *  @return The port number.
     */
    public int getPortNum(String nodeName, String ifName) {
        INode node = getNode(nodeName);
        return getPortNum(node, ifName);
    }

    /** Get a reference to a component of netlist by name.
     *  @param netlist a netlist node.
     *  @param name component name.
     *  @return The component.
     */
    public Object getComponent(INetlist netlist, String name) {
        Iterator iterator = netlist.getComponents();
        INode compNode = null;
        while (iterator.hasNext()) {
            compNode = (INode)iterator.next();
            if (compNode.getCompName(netlist).equals(name))
                return compNode.getUserObject();
        }

        return null;
    }

    /** Get a reference to a component of netlist by name.
     *  @param netlist a netlist node.
     *  @param name component name.
     *  @return The component.
     */
    public Object getComponent(Object netlist, String name) {
        return getComponent((INetlist)getNode(netlist), name);
    }

    /** Get a reference to a component of netlist by name.
     *  @param netName name of a netlist.
     *  @param name component name.
     *  @return The component.
     */
    public Object getComponent(String netName, String name) {
        INetlist netlist = getNetlist(netName);
        if (netlist == null) {
            return null;
        }
        return getComponent(netlist, name);
    }

    /** Get the node to which a node is connected through a port.
     *  @param node source node.
     *  @param port a port of the source node.
     *  @return The destination node of a connection.
     */
    public Object getConnectionDest(Object node, Object port) {
        INode inode = null;
        IPort iport = null;
        String spn = _getPortName(node, port);

        try {
            inode = getNode(node);
            iport = _findPort(inode, spn);
        } catch (Exception ex) {
            System.err.println("Network.getConnectionDest() failed");
            ex.printStackTrace();
            System.exit(1);
        }
        return getConnectionDest(inode, iport);
    }

    /** Get the node to which a node is connected through a port.
     *  @param nodeName name of source node.
     *  @param portName name of port.
     *  @return The destination node of a connection.
     */
    public Object getConnectionDest(String nodeName, String portName) {
        INode node = getNode(nodeName);
        IPort port = node.getPort(portName);
        return getConnectionDest(node, port);
    }

    /** Get the node to which a node is connected through a port.
     *  @param node source node.
     *  @param port a port of the source node.
     *  @return The destination node of a connection.
     */
    public Object getConnectionDest(INode node, IPort port) {
        Connection conn = node.getConnection(port);
        if (conn != null)
            return conn.getTarget().getUserObject();
        else
            return null;

    }

    /** Get the object that stores information about a node in the
     *  network.
     *  @param object Node of the network.
     *  @return The INode object that stores information about.
     *  the given node of this network or null if the name does not belong
     *  to any node in the network.
     */
    public INode getNode(Object object) {
        Object inst = _nodeInstances.get(object);
        //if (inst == null)
        //    throw new RuntimeException("Object is not a node of the " +
        //                 "network");
        return (INode) inst;
    }

    /** Get the object that stores information about a netlist in the
     *  network.
     *  @param object Netlist of the network.
     *  @return The INetlist object that stores information about
     *  the given netlist of this network or null if the netlist object
     *  cannot be found.
     *  @exception RuntimeException if the node found is not a netlist node
     */
    public INetlist getNetlist(Object object) {
        Object inst = _nodeInstances.get(object);
        if (inst == null)
            return null;
        //throw new RuntimeException("Object is not a node of the network");

        if (!(inst instanceof INetlist)) {
            throw new RuntimeException("Object is a node of the network, "
                    + "but it is not a netlist");
        }
        return (INetlist) inst;
    }

    /** Get the netlist to which the node is connected through port.
     *  @param node Node whose scope is set.
     *  @param port Port being referred.
     *  @return The scope.
     *  @exception RuntimeException if the node does not have such a port or
     *  the port does not implement interface scope
     *  @see #setScope(INode, IPort, INetlist)
     */
    public Object getScope(Object node, Object port) {
        INode inode = null;
        IPort iport = null;
        String spn = _getPortName(node, port);

        try {
            inode = getNode(node);
            iport = _findPort(inode, spn);
        } catch (Exception ex) {
            System.err.println("Network.getScope() failed");
            ex.printStackTrace();
            System.exit(1);
        }
        return getScope(inode, iport);
    }

    /** Get the netlist to which the node is connected through port.
     *  @param node Node whose scope is set.
     *  @param port Port being referred.
     *  @return The scope.
     *  @exception RuntimeException if the node does not have such a port or
     *  the port does not implement interface scope
     *  @see #setScope(INode, IPort, INetlist)
     */
    public Object getScope(INode node, IPort port) {
        MMType scope = getType("metamodel.lang.Scope");

        // Check that the port belongs to this node
        if (port.getContainer() != node) {
            throw new RuntimeException("getscope: port '" + port.show() +
                    "' does not belong to node '" + node.getName() +
                    "', but to node '" + port.getContainer().getName() + "'");
        }

        // Check that the port implements interface scope
        if (!port.getInterface().implementsInterface(scope)) {
            throw new RuntimeException("getscope: port '" + port.show() +
                    "' of node '" + node.getName() + "' does not implement" +
                    "interface 'metamodel.lang.Scope'");
        }

        return node.getScope(port).getUserObject();
    }

    /** Get the netlist to which the node is connected through port.
     *  @param nodeName name of node whose scope is set.
     *  @param portName name of port being referred.
     *  @return The scope.
     *  @exception RuntimeException if the node does not have such a port or
     *  the port does not implement interface scope
     *  @see #setScope(INode, IPort, INetlist)
     */
    public Object getScope(String nodeName, String portName) {
        INode node = getNode(nodeName);
        IPort port = node.getPort(portName);
        return getScope(node, port);
    }

    /** Flatten the elaborated network into a network where refined
     *  nodes and connections are replaced by their refinements.
     *  @return The netlist that contains the entire system, without
     *  refinements or redirections.
     */
    public INetlist flatten() {
        if (isFlattened())
            return _netInitiator;

        // Flatten netlists recursively, starting from the
        // net-initiator netlist.
        _flattenConnections(_netInitiator);
        _flattenNetlist(_netInitiator);

        _isFlattened = true;
        return _netInitiator;
    }

    /** Perform a 'setscope' statement in the network. Register that
     *  the 'scope' port in the node is now pointing to the netlist.
     *  @param node Node whose scope is set.
     *  @param port Port being referred.
     *  @param netlist Netlist being set.
     *  @see #getScope(Object, Object)
     *  @see #getScope(INode, IPort)
     *  @see #getScope(String, String)
     */
    public void setScope(INode node, IPort port, INetlist netlist) {
        MMType scope = getType("metamodel.lang.Scope");

        // Check that the port belongs to this node
        if (port.getContainer() != node) {
            throw new RuntimeException("setscope: port '" + port.show() +
                    "' does not belong to node '" + node.getName() +
                    "', but to node '" + port.getContainer().getName() + "'");
        }

        // Check that the port implements interface scope
        if (!port.getInterface().implementsInterface(scope)) {
            throw new RuntimeException("setscope: port '" + port.show() +
                    "' of node '" + node.getName() + "' does not implement" +
                    "interface 'metamodel.lang.Scope'");
        }

        // Check that the netlist also implements interface scope
        if (!netlist.getType().implementsInterface(scope)) {
            throw new RuntimeException("setscope: netlist '" +
                    netlist.getName() + "' of type '" +
                    netlist.getType().getName() + "' does not implement" +
                    "interface 'metamodel.lang.Scope'");
        }

        // Check that there are no other setscopes which set port
        // to another netlistobject
        Iterator iter = node.getScopePorts();
        while (iter.hasNext()) {
            Connection c = (Connection) iter.next();
            if (c.getPort() == port)
                throw new RuntimeException("setscope: scope port '" +
                        port.show() + "' is connected to netlist '" +
                        c.getTarget().getName() + "', cannot reset scope to '"
                        + netlist.getName() + "'");
        }

        // Perform the 'setscope' statement after checking that
        // everything is correct
        Connection c = new Connection(node, netlist, port, null, null);
        node.addScopeConnection(c);
        netlist.addScopeUser(c);
        addComponent(node, netlist, node.getName());
    }

    ///////////////////////////////////////////////////////////////////
    ////                         protected methods                 ////

    /** Flatten a netlist recursively. Add all components of the
     *  refined netlists to the current netlist. Remove the refinement
     *  netlists u as components of that netlist and modify
     *  connections of the network so that the last
     *  redirections/refinements are the last that are taken into
     *  account.
     *  @param netlist Netlist to be flattened.
     */
    protected void _flattenNetlist(INetlist netlist) {

        //_flattenConnections(netlist);

        List allNodes = new LinkedList();
        Iterator iter = getNodes();
        while (iter.hasNext()) {
            INode node = (INode) iter.next();
            if (node instanceof INetlist) continue;
            allNodes.add(node);
        }
        iter = allNodes.iterator();
        while (iter.hasNext()) {
            INode node = (INode) iter.next();
            String name = _removeContainers(node, netlist);
            if (!netlist.isComponent(node)) {
                addComponent(node, netlist, name + node.getName());
                //node.addCompName(netlist);
                node.addContainer(netlist);
            }
        }
    }

    /** Remove a node from all its containers except the top-level netlist.
     *  The containers no longer have this node as a component.
     *  @param node Node to be removed.
     *  @param top  Top level netlist.
     *  @return The name of the containers from which the container was
     *  removed.
     */
    protected String  _removeContainers(INode node, INetlist top) {
        Iterator containers = node.getContainers();
        String nameContainers = new String();
        while (containers.hasNext()) {
            INetlist netlist = (INetlist) containers.next();
            if (netlist == top) continue;
            nameContainers = nameContainers + netlist.getName() + ".";
            containers.remove();
            netlist.removeComponent(node);
        }

        return nameContainers;
    }

    /** Remove all connections except the connections that are not
     *  refined/redirected.
     *  @param netlist Netlist where connections should be removed.
     */
    protected void _flattenConnections(INetlist netlist) {
        Iterator iter = getNodes();
        HashSet removedConnections = new HashSet();
        while (iter.hasNext()) {
            INode node = (INode) iter.next();
            Iterator connections = node.getInConnections();
            while (connections.hasNext()) {
                Connection con = (Connection) connections.next();
                if (con.isRefined())
                    removedConnections.add(con);
            }
        }
        iter = removedConnections.iterator();
        while (iter.hasNext()) {
            Connection con = (Connection) iter.next();
            con.getTarget().removeInConnection(con);
            con.getSource().removeOutConnection(con);
        }
    }

    /** Add the information about a type.
     *  @param type Type.
     */
    public void addType(MMType type) {
        _nodeTypes.put(type.getName(),type);
    }

    /** Get the information about a type from its name.
     *  @param name Name of the type.
     *  @return The type.
     */
    public MMType getType(String name) {
        MMType type = (MMType) _nodeTypes.get(name);
        if (type == null)
            throw new RuntimeException("type " + name + " not declared");

        return type;
    }

    /** Get an iterator that will traverse all nodes in the network.
     *  @return An iterator that will generate all nodes of the network.
     */
    public Iterator getNodes() {
        return _nodeInstances.values().iterator();
    }

    /** Set the net-initiator netlist. The constructor of this netlist
     *  initializes the elaboration of the network.
     *  @param netInitiator Net-initiator netlist.
     *  @see #getNetInitiator()
     */
    public void setNetInitiator(INetlist netInitiator) {
        _netInitiator = netInitiator;
    }

    /** Get the net-initiator netlist.
     *  @return The netlist instance that builds the elaborated network.
     *  @see #setNetInitiator(INetlist)
     */
    public INetlist getNetInitiator() {
        return _netInitiator;
    }

    /** Register a new node instance in the network.
     *  @param object New node of the network.
     *  @param instance Object that contains information about this
     *  instance.
     */
    public void addNode(Object object, INode instance) {
        _nodeInstances.put(object, instance);
        _nodeNames.put(instance.getName(),instance);
    }

    /** Get the object that stores information about a node in the
     *  network from its name.
     *  @param name Name of the node in the network.
     *  @return The INode object that stores information about the
     *  given node of this network or null if the name does not belong
     *  to any node in the network.
     */
    public INode getNode(String name) {
        Object inst = _nodeNames.get(name);
        return (INode) inst;
    }

    /** Get the object that stores information about a netlist in
     *  the network from its name.
     *  @param name Name of the netlist in the network.
     *  @return The INetlist object that stores information about the
     *  given netlist of this network or null if the name does not belong
     *  to any netlist in the network.
     */
    public INetlist getNetlist(String name) {
        Object inst = _nodeNames.get(name);
        if (!(inst instanceof INetlist)) return null;
        return (INetlist) inst;
    }

    /** Test if the network has been flattened.
     *  @return true if the netlist has been flattened, false otherwise.
     */
    public boolean isFlattened() {
        if (_isFlattened)
            return true;

        // The network is flattened if the components of the net-initiator
        // meet two requirements:
        // - they have not been refined
        // - they are not of netlist type
        Iterator iter = _netInitiator.getComponents();
        _isFlattened = true;
        while (iter.hasNext() && _isFlattened) {
            INode component = (INode) iter.next();
            INetlist ref = component.getRefinement();
            _isFlattened = (ref == null) &&
                !(component instanceof INetlist);
        }
        return _isFlattened;
    }

    /** Perform a 'addcomponent' statement in the network. Register that
     *  the node is now a new component of the netlist, and that the
     *  netlist is now a new container of the component, also set the
     *  compName of the node in this netlist.
     *  @param nodeObject Node added to the netlist.
     *  @param netlistObject Netlist.
     *  @param compName The compName of node in this particular netlist.
     */
    public void addComponent(Object nodeObject, Object netlistObject,
            String compName) {
        INode node = null;
        INetlist netlist = null;
        try {
            node = getNode(nodeObject);
            netlist = getNetlist(netlistObject);
        } catch (Exception ex) {
            System.err.println("Network.addcomponent() failed");
            ex.printStackTrace();
            System.exit(1);
        }
        addComponent(node, netlist, compName);
    }

    /** Perform a 'addcomponent' statement in the network. Register that
     *  the node is now a new component of the netlist, and that the
     *  netlist is now a new container of the component, also set the
     *  compName for this node in this particular netlist.
     *  @param node The Node to be added to the netlist.
     *  @param netlist The netlist.
     *  @param compName The compName of node in this particular netlist.
     */
    public void addComponent(INode node, INetlist netlist, String compName) {
        String name;
        if (compName.length()!=0) {
            //check the uniqueness of the compName in this netlist
            Iterator comps = netlist.getComponents();
            while (comps.hasNext()) {
                INode comp = (INode) comps.next();
                String other = "";
                try {
                    other = comp.getCompName(netlist);
                } catch (Exception ex) {
                    System.err.println("Network.addcomponent() failed");
                    ex.printStackTrace();
                    System.exit(1);
                }
                if (compName.equals(other)) {
                    System.err.println("addcomponent: component name '"
                            + compName + "' for node '" + node.getName()
                            + "' is not unique in netlist '"
                            + netlist.getName() + "'.");
                    System.exit(1);
                }
            }
            name = compName;
        }
        else
            name = buildName(node.getType(), true);
        node.addContainer(netlist);
        node.addCompName(netlist, name);
        netlist.addComponent(node);
    }

    /** Perform a 'setscope' statement in the network. Register that
     *  the 'scope' port in the node is now pointing to the netlist.
     *  @param nodeObject Node whose scope is set.
     *  @param portName   Name of the scope port.
     *  @param netlistObject Netlist being set.
     *  @see #getScope(Object, Object)
     *  @see #getScope(INode, IPort)
     *  @see #getScope(String, String)
     */
    public void setScope(Object nodeObject, String portName,
            Object netlistObject) {
        INode node = null;
        IPort port = null;
        INetlist netlist = null;
        try {
            node = getNode(nodeObject);
            port = _findPort(node, portName);
            netlist = getNetlist(netlistObject);
        } catch (Exception ex) {
            System.err.println("Network.setscope() failed");
            ex.printStackTrace();
            System.exit(1);
        }
        setScope(node, port, netlist);
    }

    /** Perform a 'setscope' statement in the network. Register that
     *  the 'scope' port in the node is now pointing to the netlist.
     *  @param node Node whose scope is set.
     *  @param portName Name of the scope port.
     *  @param netlist Netlist being set.
     *  @see #getScope(Object, Object)
     *  @see #getScope(INode, IPort)
     *  @see #getScope(String, String)
     */
    public void setScope(INode node, String portName, INetlist netlist) {
        IPort port = _findPort(node, portName);
        setScope(node, port, netlist);
    }

    /** Perform a 'connect' statement in the network. Register that
     *  the two objects are now connected by a given port.
     *  @param srcObject Node whose port is connected to the target.
     *  @param srcPort Port connected to the target.
     *  @param destObject Destination of the connection.
     */
    public void connect(Object srcObject, Object srcPort,
            Object destObject) {
        INode src = null;
        IPort port = null;
        INode dst = null;
        String spn = _getPortName(srcObject, srcPort);

        try {
            src = getNode(srcObject);
            port = _findPort(src, spn);
            dst = getNode(destObject);
        } catch (Exception ex) {
            System.err.println("Network.connect() failed");
            ex.printStackTrace();
            System.exit(1);
        }
        connect(src, port, dst);
    }

    /** Perform a 'connect' statement in the network. Register that
     *  the two objects are now connected by a given port.
     *  @param src Node whose port is connected to the target.
     *  @param srcPortName Port connected to the target.
     *  @param dest Destination of the connection.
     */
    public void connect(INode src, String srcPortName, INode dest) {
        IPort srcPort = src.getPort(srcPortName);
        connect(src, srcPort, dest);
    }

    /** Perform a 'connect' statement in the network. Register that
     *  the two objects are now connected by a given port.
     *  @param src Node whose port is connected to the target.
     *  @param srcPort Port connected to the target.
     *  @param dest Destination of the connection.
     */
    public void connect(INode src, IPort srcPort, INode dest) {
        // Check that the port belongs to the source node
        if (srcPort.getContainer() != src) {
            throw new RuntimeException("connect: port '" + srcPort.show()
                    + "' does not belong to node '" + src.getName()
                    + "', but to node '" + srcPort.getContainer().getName()
                    + "'");
        }

        // Check that the target node implements the interface
        // defined by the port
        if (!dest.getType().implementsInterface(srcPort.getInterface())) {
            throw new RuntimeException("connect: target '" + dest.getName()
                    + "' of type '" + dest.getType().getName() + "' does not"
                    + "implement the interface '"
                    + srcPort.getInterface().getName() + "' of the port '"
                    + srcPort.getName() + "'");
        }

        // Check that there aren't other connections using this port
        Connection c = src.getConnection(srcPort);
        if (c != null) {
            throw new RuntimeException("connect: port '" + srcPort.show()
                    + "' already connected to node '" + c.getTarget().getName()
                    + "', so it cannot be connected"
                    + " again to another node, '" + dest.getName() + "'");
        }

        // Perform the connection after checking correctness
        c = new Connection(src, dest, srcPort, null, null);
        src.addOutConnection(c);
        dest.addInConnection(c);

        /*        try {
                  Object srcObj = src.getUserObject();
                  Object destObj = dest.getUserObject();
                  Class cls = srcObj.getClass();
                  boolean flag = false;
                  Field f = null;
                  Field g = null;

                  do {
                  try {
                  f = cls.getDeclaredField(srcPort.getName());
                  flag = true;
                  } catch (NoSuchFieldException e) {
                  cls = cls.getSuperclass();
                  if (cls.getName().equals("java.lang.Object")) throw e;
                  }
                  } while (flag==false);

                  f.setAccessible(true);
                  f.set(srcObj, destObj);
                  } catch (NoSuchFieldException ex) {
                  System.err.println("Network.connect() failed");
                  System.err.println("In process " + src.getName()
                  + ", port '" + srcPort.getName()
                  + "' does not exist.");
                  ex.printStackTrace();
                  System.exit(1);
                  } catch (IllegalAccessException e) {
                  System.err.println("Network.connect() failed");
                  ex.printStackTrace();
                  System.exit(1);
                  }
        */
    }

    /** Perform a 'refine' statement in the network. Register that the
     *  netlist now refines the given node.
     *  @param nodeObject Node being refined.
     *  @param netlistObject Netlist describing the refinement.
     */
    public void refine(Object nodeObject, Object netlistObject) {
        INode node = null;
        INetlist netlist = null;
        try {
            node = getNode(nodeObject);
            netlist = getNetlist(netlistObject);
        } catch (Exception ex) {
            System.err.println("Network.refine() failed");
            ex.printStackTrace();
            System.exit(1);
        }
        refine(node, netlist);
    }

    /** Perform a 'refine' statement in the network. Register that the
     *  netlist now refines the given node.
     *  @param node Node being refined.
     *  @param netlist Netlist describing the refinement.
     */
    public void refine(INode node, INetlist netlist) {
        // Check that the node is not refined several times
        if (node.isRefined()) {
            throw new RuntimeException("refine: node '" + node.getName() +
                    "' already refined by netlist '" +
                    node.getRefinement().getName() + "', so it cannot be refined"
                    + "again by netlist '" + netlist.getName() + "'");
        }

        // Register the refinement after checking correctness
        _isFlattened = false;
        node.setRefinement(netlist);
        netlist.addRefinedNode(node);
    }


    /** Perform a 'refineconnect' statement in the network. Register that
     *  the target of the connection has changed.
     *  @param netlistObject Netlist defining the refinement.
     *  @param srcObject Source of the connection.
     *  @param portRef Port defining the connection being refined.
     *  @param componentObject New destination of the connection.
     */
    public void refineConnect(Object netlistObject, Object srcObject,
            Object portRef, Object componentObject) {
        INetlist netlist = null;
        INode src  = null;
        IPort port = null;
        INode dst  = null;
        String pn = _getPortName(srcObject, portRef);
        try {
            netlist = getNetlist(netlistObject);
            src = getNode(srcObject);
            port = _findPort(src, pn);
            dst = getNode(componentObject);
        } catch (Exception e) {
            System.err.println("refineconnect: " + e.toString());
            System.exit(1);
        }
        refineConnect(netlist, src, port, dst);
    }

    /** Perform a 'refineconnect' statement in the network. Register that
     *  the target of the connection has changed.
     *  @param netlist Netlist defining the refinement.
     *  @param src Source of the connection.
     *  @param portName Port defining the connection being refined.
     *  @param component New destination of the connection.
     */
    public void refineConnect(INetlist netlist, INode src, String portName,
            INode component) {
        IPort port = _findPort(src, portName);
        refineConnect(netlist, src, port, component);
    }

    /** Perform a 'refineconnect' statement in the network. Register that
     *  the target of the connection has changed.
     *  @param netlist Netlist defining the refinement.
     *  @param src Source of the connection.
     *  @param port Port defining the connection being refined.
     *  @param component New destination of the connection.
     */
    public void refineConnect(INetlist netlist, INode src, IPort port,
            INode component) {
        // Check that the source port belongs to the source object
        if (port.getContainer() != src) {
            throw new RuntimeException("refineconnect: port '" + port.show() +
                    "' does not belong to node '" + src.getName() +
                    "', but to node '" + port.getContainer().getName() + "'");
        }

        // Check that component implements the interface in the port
        if (!component.getType().implementsInterface(port.getInterface())) {
            throw new RuntimeException("refineconnect: component '" +
                    component.getName() + "' of type '" +
                    component.getType().getName() + "' does not implement the " +
                    "interface '" + port.getInterface().getName() + "' of the " +
                    "port '" + port.getName() + "'");
        }

        // Check that component is a component of the netlist
        if (!netlist.isComponent(component)) {
            throw new RuntimeException("refineconnect: node '" +
                    component.getName() + "' is not a component of netlist '" +
                    netlist.getName() + "' although it is required by the " +
                    "refineconnect statement");
        }

        // Check that there is one connection from the source node that
        // uses the port. This will be the connection to be refined
        Connection last = src.getConnection(port);
        if (last == null) {
            throw new RuntimeException("refineconnect: connection to be "+
                    "refined does not exist, there is no connection from node '"
                    + src.getName() + "' using port '" + port.show() + "'");
        }

        // Build the refined connection
        _isFlattened = false;
        Connection refined = new Connection(src, component, port, null, last);
        component.addInConnection(refined);
        src.addOutConnection(refined);
    }

    /** Perform a 'redirectconnect' statement in the network. Register
     *  the source of the connection has changed.
     *  @param netlistObject Netlist defining the refinement.
     *  @param orgObject Old source of the connection.
     *  @param orgPortRef Old port that performs the connection.
     *  @param componentObject New source of the connection.
     *  @param newPortRef New port of the connection.
     */
    public void redirectConnect(Object netlistObject, Object orgObject,
            Object orgPortRef, Object componentObject, Object newPortRef) {

        INetlist netlist = null;
        INode origNode = null, newNode = null;
        IPort origPort = null, newPort = null;
        String orgPortName = _getPortName(orgObject, orgPortRef);
        String newPortName= _getPortName(componentObject, newPortRef);
        try {
            netlist =  getNetlist(netlistObject);
            origNode = getNode(orgObject);
            origPort = _findPort(origNode, orgPortName);
            newNode =  getNode(componentObject);
            newPort =  _findPort(newNode, newPortName);
        } catch (Exception e) {
            System.err.println("redirectconnect: " + e.toString());
            System.exit(1);
        }
        redirectConnect(netlist, origNode, origPort, newNode, newPort);
    }

    /** Perform a 'redirectconnect' statement in the network. Register
     *  the source of the connection has changed.
     *  @param netlist Netlist defining the refinement.
     *  @param org Old source of the connection.
     *  @param orgPortName Old port that performs the connection.
     *  @param component New source of the connection.
     newPort =  _findPort(newNode, newPortName);
     *  @param newPortName New port of the connection.
     */
    public void redirectConnect(INetlist netlist, INode org,
            String orgPortName, INode component, String newPortName) {
        IPort orgPort = _findPort(org, orgPortName);
        IPort newPort = _findPort(component, newPortName);
        redirectConnect(netlist, org, orgPort, component, newPort);
    }

    /** Perform a 'redirectconnect' statement in the network. Register
     *  the source of the connection has changed.
     *  @param netlist Netlist defining the refinement.
     *  @param org Old source of the connection.
     *  @param orgPort Old port that performs the connection.
     *  @param component New source of the connection.
     *  @param newPort New port of the connection.
     */
    public void redirectConnect(INetlist netlist, INode org, IPort orgPort,
            INode component, IPort newPort) {
        // Check that the instances of ports passed as parameters belong
        // to the objects passed as parameters
        if (orgPort.getContainer() != org) {
            throw new RuntimeException("redirectconnect: port '" +
                    orgPort.show() + "' of the original connection does not " +
                    "belong to node '" + org.getName() + "', but to node '" +
                    orgPort.getContainer().getName() + "'");
        }
        if (newPort.getContainer() != component) {
            throw new RuntimeException("redirectconnect: port '" +
                    newPort.show() + "' to be used as the replacement port " +
                    "does not belong to the node '" + component.getName() +
                    "but to node '" + newPort.getContainer().getName() + "'");
        }

        // Check that the interface implemented by the original port is a
        // subinterface of the interface implemented by the new port
        MMType orgIf = orgPort.getInterface();
        MMType newIf = newPort.getInterface();
        if (!orgIf.implementsInterface(newIf)) {
            throw new RuntimeException("redirectconnect: port '" +
                    orgPort.show() + "' of the original connection does not" +
                    " implement the interface '" + newPort.show() + "' of the" +
                    "redirected connection");
        }

        // Check that component is a component of the netlist
        if (!netlist.isComponent(component)) {
            throw new RuntimeException("redirectconnect: node '" +
                    component.getName() + "' is not a component of netlist '" +
                    netlist.getName() + "' although it is required by the " +
                    "redirectconnect statement");
        }

        // Check that there is a valid connection from the original node
        // using the original port. Also check that there are no connections
        // from the component node using the new port
        Connection last = org.getConnection(orgPort);
        if (last == null) {
            throw new RuntimeException("redirectconnect: port '" +
                    orgPort.show() + "' of node '" + org.getName() +
                    "' is not used by any connection, so we cannot redirect " +
                    "this connection");
        }
        Connection other = component.getConnection(newPort);
        if (other != null) {
            throw new RuntimeException("redirectconnect: port '" +
                    newPort.show() + "' of node '" + component.getName() +
                    "' is already in use, so we cannot redirect a connection " +
                    " to this port");
        }

        // Perform the redirection after checking correctness
        _isFlattened = false;
        Connection redirect =
            new Connection(component, last.getTarget(), newPort, null, last);
        last.getTarget().addInConnection(redirect);
        component.addOutConnection(redirect);
    }


    /** Perform a 'getconnectionnum' statement. Return the number of
     *  connections to a given node that implement a given interface.
     *  @param mediumObject Node.
     *  @param ifName Name of the interface.
     *  @return The number of connections.
     */
    public int getConnectionNum(Object mediumObject, String ifName) {
        INode medium = null;
        MMType intf = null;
        try {
            medium = getNode(mediumObject);
            intf = getType(ifName);
        } catch (Exception e) {
            System.err.println("getconnectionnum: " + e.toString());
            System.exit(1);
        }
        return getConnectionNum(medium, intf);
    }

    /** Perform a 'getconnectionnum' statement. Return the number of
     *  connections to a given node that implement a given interface.
     *  @param medium Node.
     *  @param ifName Name of the interface.
     *  @return The number of connections.
     */
    public int getConnectionNum(INode medium, String ifName) {
        MMType intf = getType(ifName);
        return getConnectionNum(medium, intf);
    }

    /** Perform a 'getconnectionnum' statement. Return the number of
     *  connections to a given node that implement a given interface.
     *  @param medium Node.
     *  @param intf The interface.
     *  @return The number of connections.
     */
    public int getConnectionNum(INode medium, MMType intf) {
        // Check that the object is a medium
        MMType mediumType = medium.getType();
        if ((mediumType.getKind() != MMType.MEDIUM) &&
                (mediumType.getKind() != MMType.STATEMEDIUM)) {
            throw new RuntimeException("getconnectionnum: node '" +
                    medium.getName() + "' used as the parameter is not a " +
                    "medium, as it has type '" + medium.getType().getName() +
                    "'");
        }

        // Check that the type is an interface
        if (intf.getKind() != MMType.INTERFACE) {
            throw new RuntimeException("getconnectionnum: type with name '" +
                    intf.getName() + "' is not an interface");
        }

        // Check that the object implements the interface
        if (!mediumType.implementsInterface(intf)) {
            throw new RuntimeException("getconnectionnum: medium '" +
                    medium.getName() + "' does not implement interface '" +
                    intf.getName() + "'");
        }

        // Count the number of connections
        return medium.getNumUsers(intf);
    }

    /** Perform a 'getnthconnectionsrc' statement. Get the target of the
     *  n-th connection to a given node that implements a given interface.
     *  @param mediumObject Node.
     *  @param ifName Name of the interface.
     *  @param n Index of the connection being accessed.
     *  @return The target of the n-th connection to the node implementing
     *  the interface with the given name.
     */
    public Object getNthConnectionSrc(Object mediumObject, String ifName,
            int n) {
        INode medium = null;
        MMType intf = null;
        Connection c = null;
        try {
            medium = getNode(mediumObject);
            intf = getType(ifName);
            c = medium.getNthUser(intf, n);
        } catch (Exception e) {
            System.err.println("getnthconnectionsrc: " + e.toString());
            System.exit(1);
        }
        return getNthConnectionSrc(medium, intf, n);
    }

    /** Perform a 'getnthconnectionsrc' statement. Get the target of the
     *  n-th connection to a given node that implements a given interface.
     *  @param medium Node.
     *  @param ifName Name of the interface.
     *  @param n Index of the connection being accessed.
     *  @return The target of the n-th connection to the node implementing
     *  the interface with the given name.
     */
    public Object getNthConnectionSrc(INode medium, String ifName, int n) {
        MMType intf = getType(ifName);
        return getNthConnectionSrc(medium, intf, n);
    }

    /** Perform a 'getnthconnectionsrc' statement. Get the target of the
     *  n-th connection to a given node that implements a given interface.
     *  @param medium Node.
     *  @param intf Interface.
     *  @param n Index of the connection being accessed.
     *  @return The target of the n-th connection to the node implementing
     *  the interface with the given name.
     */
    public Object getNthConnectionSrc(INode medium, MMType intf, int n) {
        int max = 0;
        Connection c = null;
        try {
            max = getConnectionNum(medium, intf);
            if (max < n) {
                throw new RuntimeException("medium '" + medium.getName() +
                        "'  does not have " + n + " connections implementing" +
                        " interface '" + intf.getName() + "', it only has " +
                        max);
            }
            c = medium.getNthUser(intf, n);
        } catch (Exception e) {
            throw new RuntimeException("getnthconnectionsrc: " +
                    e.getMessage());
        }
        return c.getSource().getUserObject();
    }

    /** Perform a 'getnthconnectionport' statement. Get the port of the
     *  n-th connection to a given node that implements a given interface.
     *  @param mediumObject Node.
     *  @param ifName Name of the interface.
     *  @param n Index of the connection being accessed.
     *  @return The port used in the n-th connection to the node implementing
     *  the interface with the given name.
     */
    public Object getNthConnectionPort(Object mediumObject, String ifName,
            int n) {
        INode medium = null;
        MMType intf = null;
        Connection c = null;
        try {
            medium = getNode(mediumObject);
            intf = getType(ifName);
            c = medium.getNthUser(intf, n);
        } catch (Exception e) {
            System.err.println("getnthconnectionport: " + e.toString());
            System.exit(1);
        }

        try {
            Object obj = c.getSource().getUserObject();
            Class cls = obj.getClass();
            IPort iport = c.getPort();
            boolean flag = false;
            Field f = null;

            do {
                try {
                    f = cls.getDeclaredField(iport.getName());
                    flag = true;
                } catch (NoSuchFieldException e) {
                    cls = cls.getSuperclass();
                    if (cls.getName().equals("java.lang.Object")) throw e;
                }
            } while (flag==false);

            f.setAccessible(true);
            if (iport instanceof IPortElem) {
                return Array.get(f.get(obj), ((IPortElem)iport).getElemPortIndex());
            } else {
                return f.get(obj);
            }
        } catch (NoSuchFieldException e) {
            System.err.println("getnthconnectionport: " + e.toString());
            System.err.println("In medium "+medium.getName() +
                    ", there is no port '"+c.getPort().getName() +
                    "' with interface type "+ifName+".");
            System.exit(1);
        } catch (IllegalAccessException e) {
            System.err.println("getnthconnectionport: " + e.toString());
            System.exit(1);
        }

        return null;
    }

    /** Allocate an array of ports with a given run-time dimension.
     *  @param nodeObject Node where the array of ports is instantiated.
     *  @param portName Name of the array of ports.
     *  @param idxString String describing the size of the array of
     *  ports. The String must be defined as "[x][y][z]..." where x,
     *  y and z are integer constants.
     *  @exception RuntimeException if the node does not have an array
     *  with that name, or if the port is not an array of ports, or
     *  if the size of the index does not match with the number of
     *  dimensions of the array, or if any of the indexes is <= 0.
     */
    public void instantiatePortArray(Object nodeObject, String portName,
            String idxString) {
        INode node = null;
        IPort port = null;
        int idx[]  = null;
        try {
            node = getNode(nodeObject);
            port = node.getPort(portName);
            if (port == null)
                throw new RuntimeException("Node '" + node.getName() +
                        "' does not have any port called '" + portName + "'");
            if (!(port instanceof IPortArray)) {
                throw new RuntimeException("Port '" + portName +
                        "' of node '" + node.getName() + "' is not an array");
            }
            // Parse the idxString
            StringTokenizer token = new StringTokenizer(idxString,"]");
            int count = token.countTokens();
            if (port.getDecl().getDims() != count)
                throw new RuntimeException("Port '" + port.show() +
                        "' does not have the same dimensions as '" + idx);
            idx = new int[count];
            int i = 0;
            while (token.hasMoreTokens()) {
                String number = token.nextToken().substring(1);
                idx[i] = Integer.parseInt(number);
                if (idx[i] < 0)
                    throw new RuntimeException("Index '" + idx[i] +
                            "' used to allocate port '" + portName +
                            "' of node '" + node.getName() + "' is invalid");
                i++;
            }
            IPortArray array = (IPortArray) port;
            array.allocate(idx);
        } catch (Exception e) {
            System.err.println("Allocating array of ports: " + e.toString());
            System.exit(1);
        }
    }

    /** Get a String with information about all nodes in the
     *  network. The amount of information provided this way can be
     *  very large, so the usage of show() in each node is
     *  recommended.
     *  @return A String with information about the nodes in the network.
     */
    public String show() {
        StringBuffer results = new StringBuffer("Top-Level netlist is "
                + (_netInitiator == null ? "null" : _netInitiator.getName())
                + "\n");

        Iterator nodes = _nodeInstances.values().iterator();
        while (nodes.hasNext()) {
            INode node = (INode) nodes.next();
            //results = results + "\n" + node.show() + "\n";
            results.append("\n" + node.show(_netInitiator) + "\n");
        }

        results.append("\n### List of annotations ###\n");
        Enumeration events = _annotations.keys();
        while (events.hasMoreElements()) {
            Event event = (Event)events.nextElement();
            LinkedList vars = (LinkedList)_annotations.get(event);

            results.append("   o " + event.show());
            Iterator eventVariables = vars.iterator();
            while (eventVariables.hasNext()) {
                String eventVariable = (String)eventVariables.next();
                results.append(" " + eventVariable);
            }
            results.append("\n");
        }

        results.append( "\n### List of variables ###\n");
        Iterator variables = _variables.iterator();
        while (variables.hasNext()) {
            String variable = (String) variables.next();
            results.append(variable + "\t");
        }
        results.append("\n");

        return results.toString();
    }

    /** Save the information about a network to a file. Use Java Object
     *  Serialization to store the objects to disk.
     *  @param fileName Name of the file.
     */
    public void save(String fileName) {
        try {
            FileOutputStream ostream = new FileOutputStream(fileName);
            ObjectOutputStream out = new ObjectOutputStream(ostream);
            out.writeObject(Network.net);
            out.flush();
            //ostream.close();
            out.close();
        } catch (Exception e) {
            throw new RuntimeException("Error writing network to path '"
                    + fileName + "': " + e.toString());
        }
    }

    /** Restore the information about a network from a file. The information
     *  will have been stored using the Serialization procedure offered by
     *  Java.
     *  @param fileName Name of the file.
     *  @return The network.
     */
    public static Network restore(String fileName) {
        Network n = null;
        try {
            FileInputStream istream = new FileInputStream(fileName);
            ObjectInputStream in = new ObjectInputStream(istream);
            n = (Network) in.readObject();
            istream.close();
        } catch (Exception e) {
            throw new RuntimeException("Error reading network from path '"
                    + fileName + "':" + e.toString());
        }
        return n;
    }

    /** Restore the information about a network from an ObjectInputStream.
     *  Close the stream after reading. Catch all possible exceptions.
     *  @param ois The ObjectInputStream.
     *  @return The network.
     */
    public static Network restore(ObjectInputStream ois) {
        // FIXME: Finish the implementation of this class
        Network network = null;
        try {
            network = (Network) ois.readObject();
        } catch (Throwable throwable) {
            throw new RuntimeException("Error reading ObjectInputStream.",
                    throwable);
        }
        return network;
    }

    /** Get a new name for a node instance. Check that the name is not
     *  being used by any other node instance in the network.
     *  @param type The type of the node instance.
     *  @param forCompName Build instName for a node if forCompName
     *         equals true, otherwise generate an instName.
     *  @return The name for the node instance.
     */
    public String buildName(MMType type, boolean forCompName) {
        String name = type.getName();
        name = StringManip.partAfterLast(name, '.');

        if (forCompName) name = name+"_Comp_";

        int count = 0;
        if (_instanceCount.containsKey(name)) {
            Integer val = (Integer) _instanceCount.get(name);
            count = val.intValue() + 1;
        }
        _instanceCount.put(name, new Integer(count));
        return name +  count;
    }

    /** Add an annotation to the annotation list of the network.
     *  @param e the event associated with the annotation.
     *  @param v the variable name of the annotation.
     */
    public void addAnnotation(Event e, String v)
    {
        if (e == null)
            return;
        LinkedList vars;
        addVariable(v);
        Enumeration events = _annotations.keys();
        while (events.hasMoreElements()) {
            Event current = (Event)events.nextElement();
            if (current.equals(e))
                {
                    vars = (LinkedList)_annotations.get(current);
                    Iterator iter = vars.iterator();
                    while (iter.hasNext())
                        {
                            String var = (String)iter.next();
                            if (var.equals(v))
                                return;
                        }
                    vars.addLast(v);
                    return;
                }
        }

        vars = new LinkedList();
        vars.addLast(v);
        _annotations.put(e, vars);
    }

    /** Get a list of annotations in the network.
     *  @return A list of mappings from event objects to lists of variables.
     */
    public Hashtable getAnnotations()
    {
        return _annotations;
    }

    /** Add a constraint to the constraint list of the network.
     *  @param con the constraint that needs to be added.
     */
    public void addConstraint(Constraint con)
    {
        if (con == null)
            return;
        Iterator iter = _constraints.iterator();
        while (iter.hasNext())
            {
                Constraint c = (Constraint)iter.next();
                if (c.getSource().equals(con.getSource()))
                    return;
            }
        _constraints.addFirst(con);
    }

    /** Get a list of constraints in the network.
     *  @return A list of constraints in the network.
     */
    public LinkedList getConstraints()
    {
        return _constraints;
    }

    /** Add a variable name to the variable set.
     *  @param var the variable name that is added.
     */
    public void addVariable(String var)
    {
        if (var == null)
            return;
        _variables.add(var);
    }

    /** Get a set of variable names that appear in the annotation list.
     *  @return A set of variable names
     */
    public LinkedHashSet getVariables()
    {
        return _variables;
    }

    /** Generate constraint checker files.
     */
    public void generateCheckers()
    {
        String annotStr = new String("annotation: event");
        String traceStr = new String("trace: \"%s");
        Iterator iter = _variables.iterator();
        while (iter.hasNext()) {
            String var = (String)iter.next();
            annotStr += " " + var;
            traceStr += " %d";
        }

        int counter = 0;
        String fileName;
        iter = _constraints.iterator();
        while (iter.hasNext()) {
            Constraint constr = (Constraint)iter.next();
            if (constr.getKind() == Constraint.LOC) {
                fileName = ".constr" + (counter++) + ".loc";
                try {
                    FileWriter locFile = new FileWriter(fileName);
                    locFile.write("loc: " + constr.getFormula() + "\n");
                    locFile.write(annotStr + "\n");
                    locFile.write(traceStr + "\"");
                    locFile.flush();
                    locFile.close();
                } catch (Exception ex) {
                    throw new RuntimeException("Error generating checker "
                            + "to path '" + fileName + "'", ex);
                }
            }
        }
        return;
    }

    ///////////////////////////////////////////////////////////////////
    ////                        protected variables                ////

    /** Netlist that builds the elaborated network. */
    protected INetlist _netInitiator;

    /** Flag indicating if the network has been flattened. A flattened
     *  network is a single netlist, where refined nodes have been replaced
     *  by their refinement and connections are not redirected. */
    protected boolean _isFlattened = false;

    /** Mapping from the node instances in the network to the objects
     *  that keep information about the connection/refinements of that
     *  instance. */
    protected Hashtable _nodeInstances = new Hashtable();

    /** Mapping from the name of a node to INode object that has
     *  information about this instance. */
    protected Hashtable _nodeNames = new Hashtable();

    /** Mapping from the name of a type to the number of instances of
     *  nodes with that name. */
    protected Hashtable _instanceCount = new Hashtable();

    /** Mapping from the names of the type objects containing
     *  information about the types.
     */
    protected Hashtable _nodeTypes = new Hashtable();

    /** Mapping from event objects containing a variable. */
    protected Hashtable _annotations = new Hashtable();

    /** A list of constraints in the network. */
    protected LinkedList _constraints = new LinkedList();

    /** A set of variable names that appear in the annotation list. */
    protected LinkedHashSet _variables = new LinkedHashSet();

    /** Remove all refined versions of a given connection. Leave only
     *  the latest version of this connection.
     *  @param con Connection being flattened.
     */
    protected void _refineConnection(Connection con) {
        Connection last = con.finalConnection();
        Connection current = last.getRefinedConnection();
        while (current != null) {
            current.getTarget().removeInConnection(current);
            current.getSource().removeOutConnection(current);
            current = current.getRefinedConnection();
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                          private methods                  ////

    /** Get the instance of a port that is referenced by a String.
     *  The String has the form "portName" or "portName[x][y]", where
     *  x,y... are integer constants.
     *  @param node Instance of the node where the port is declared.
     *  @param port String that defines the port.
     *  @return The instance of a port that is referenced by a String.
     *  @exception RuntimeException if the node instance does not have a
     *  port with that name, or if the port should be an array and
     *  is not accessed with an index or viceversa.
     *  @exception IndexOutOfBoundsException if the port is an array but
     *  it does not have an element with this index.
     */
    private IPort _findPort(INode node, String port) {
        StringTokenizer tokenizer = new StringTokenizer(port,"[");
        int count = tokenizer.countTokens();
        String portName  = tokenizer.nextToken();
        if (count == 1) {
            // The port should be scalar, because it is not
            // specified using an index
            IPort thePort = node.getPort(portName);
            if (thePort == null)
                throw new RuntimeException("Node '" + node.getName() +
                        "' does not have a port called '" + portName + "'");
            if (thePort instanceof IPortArray)
                throw new RuntimeException("Port '" + portName +
                        "' of node '" + node.getName() + "' is an array" +
                        " of ports, so it cannot be used as an scalar.");
            return thePort;
        } else {
            // The port should be an array of ports
            int idx[] = new int[count-1];
            int i = 0;
            while (tokenizer.hasMoreTokens()) {
                String number = tokenizer.nextToken();
                number = number.substring(0,number.length()-1);
                idx[i++] = Integer.parseInt(number);
            }

            // Get the element of the port in the node
            IPort thePort = node.getPort(portName,idx);
            return thePort;
        }
    }

    /** Get the name of a port defined in an array of object
     *  @param obj The array object
     *  @param reference to the port
     *  @return The name of the port in obj,
     *                                 or null if the port does not exist in the array
     */
    private String _getPortIndexFromArray(Object obj, Object port) {
        Class cls = obj.getClass();
        int length = Array.getLength(obj);
        for (int i=0; i<length; i++) {
            Object ithElem = Array.get(obj, i);
            if (ithElem == null) continue;
            if (ithElem.getClass().isArray()) {
                String pn = _getPortIndexFromArray(ithElem, port);
                if (pn != null) return ("["+i+"]"+pn);
            } else {
                if (ithElem == port) return "["+i+"]";
            }
        }//end for

        return null;
    }

    /** Get the name of a port defined in an object
     *  @param obj The object
     *  @param port String or a reference to the port
     *  @return The name of the port in obj
     *  @exception RuntimeException if port does exist in obj
     */
    private String _getPortName(Object obj, Object port) {
        if (port instanceof String) return (String)port;
        Class cls = obj.getClass();

        while (cls!=null && !cls.getName().equals("java.lang.Object")) {
            Field f[] = cls.getDeclaredFields();
            AccessibleObject.setAccessible(f, true);
            for (int i=0; i<f.length; i++) {
                try {
                    Object ithField = f[i].get(obj);
                    if (ithField == null)
                        continue;
                    else if (ithField  == port)
                        return f[i].getName();
                    else
                        if (ithField.getClass().isArray()) {
                            String pn = _getPortIndexFromArray(ithField, port);
                            if (pn!=null) return (f[i].getName()+pn);
                        }
                } catch (IllegalAccessException e) {
                    throw new RuntimeException("Cannot access object "+
                            cls.getName()+" to determine its ports.");
                }
            }
            cls = cls.getSuperclass();
        }
        throw new RuntimeException(obj.getClass().getName()+"does not have a " +
                port.getClass().getName()+ " typed port");
    }
}
